# -*- coding: utf-8 -*-
import logging

from .openbabel import OpenBabelMixin
from .smiles import SMILESMixin

logger = logging.getLogger(__name__)


class _Template(SMILESMixin, OpenBabelMixin, object):
    """:meta public:
    A class providing the API for templates.

    There are two types of templates:

    simple
        Which have a category and name, which together are unique.
    full
        Which have a category and name, which together are unique,
        and also a reference to a configuration that
        describes the atoms, bonds, and other properties of the
        template.

    Parameters
    ----------
    system_db : SystemDB
        The system database that this template belongs to.
    tid : int
        The id of the template in the template table.
    logger : logging.Logger = logger
        A logger to use in place of the one from this module.
    """

    def __init__(self, system_db, tid, logger=logger):
        self._system_db = system_db
        self._id = tid
        self._logger = logger

        # Cache for performance:
        self._is_full = None
        self._configuration = None
        self._configuration_id = None

    @property
    def category(self):
        """The category of this template."""
        sql = "SELECT category FROM template WHERE id = ?"
        self.cursor.execute(sql, (self.id,))
        return self.cursor.fetchone()[0]

    @property
    def configuration(self):
        """The configuration for a full template.

        Raises
        ------
        TypeError
            If not a full template, so there is no template configuration.

        Returns
        -------
        _Configuration
            The Configuration object.
        """
        if not self.is_full:
            raise TypeError("Not a full template")

        if self._configuration is None:
            cid = self.configuration_id
            if cid is None:
                return None
            self._configuration = self.system_db.get_configuration(cid)
        return self._configuration

    @property
    def configuration_id(self):
        """The configuration id for a full template.

        Raises
        ------
        TypeError
            If not a full template.

        Returns
        -------
        int
            The id of the configuration in the configuration table.
        """
        if not self.is_full:
            raise TypeError("Not a full template")

        if self._configuration_id is None:
            sql = "SELECT configuration FROM template WHERE id = ?"
            self.cursor.execute(sql, (self.id,))
            self._configuration_id = self.cursor.fetchone()[0]
        return self._configuration_id

    @property
    def atoms(self):
        """The atoms for this template.

        Raises
        ------
        TypeError
            If not a full template.

        Returns
        -------
        _Atoms
            The atoms for the template configuration.
        """

        return self.configuration.atoms

    @property
    def bonds(self):
        """The bonds for this template.

        Raises
        ------
        TypeError
            If not a full template.

        Returns
        -------
        _Bonds
            The bonds for the template configuration.
        """

        return self.configuration.bonds

    @property
    def cell(self):
        """The cell for this template.

        Raises
        ------
        TypeError
            If not a full template or is not periodic.

        Returns
        -------
        _Cell
            The cell for the template configuration.
        """

        return self.configuration.cell

    @property
    def coordinate_system(self):
        """The coordinate system for this template.

        Raises
        ------
        TypeError
            If not a full template.

        Returns
        -------
        str
            The coordinate system for the template configuration.
        """

        return self.configuration.coordinate_system

    @property
    def cursor(self):
        """The a cursor for the database."""
        return self.system_db.cursor

    @property
    def db(self):
        """The database connection."""
        return self.system_db.db

    @property
    def density(self):
        """The density for this template.

        Raises
        ------
        TypeError
            If not a full template or is not periodic.

        Returns
        -------
        float
            The density for the template configuration.
        """

        return self.configuration.density

    @property
    def formula(self):
        """The chemical formula for this template.

        Raises
        ------
        TypeError
            If not a full template.

        Returns
        -------
        tuple(str, str, int)
            The chemical formula, empirical formula and Z.
        """

        return self.configuration.formula

    @property
    def is_full(self):
        """Whether this has a template configuration, i.e. is a full template."""
        if self._is_full is None:
            if self._configuration_id is None:
                sql = "SELECT configuration FROM template WHERE id = ?"
                self.cursor.execute(sql, (self.id,))
                cid = self.cursor.fetchone()[0]
                self._is_full = cid is not None
            else:
                self._is_full = True
        return self._is_full

    @property
    def id(self):
        """The id for this template in the template table."""
        return self._id

    @property
    def mass(self):
        """The atomic mass of this template.

        Raises
        ------
        TypeError
            If not a full template.

        Returns
        -------
        float
            The atomic mass of the template configuration.
        """

        return self.configuration.mass

    @property
    def name(self):
        """The name of this template."""
        sql = "SELECT name FROM template WHERE id = ?"
        self.cursor.execute(sql, (self.id,))
        return self.cursor.fetchone()[0]

    @property
    def n_atoms(self):
        """The number of atoms for this template.

        Raises
        ------
        TypeError
            If not a full template.

        Returns
        -------
        int
            The number of atoms for the template configuration.
        """

        return self.configuration.n_atoms

    @property
    def n_bonds(self):
        """The number of bonds for this template.

        Raises
        ------
        TypeError
            If not a full template.

        Returns
        -------
        int
            The number of bonds for the template configuration.
        """

        return self.configuration.n_bonds

    @property
    def periodicity(self):
        """The periodicity of this template.

        Raises
        ------
        TypeError
            If not a full template.

        Returns
        -------
        int
            The periodicity of the template configuration.
        """

        return self.configuration.periodicity

    @property
    def symmetry(self):
        """The symmetry of this template.

        Raises
        ------
        TypeError
            If not a full template.

        Returns
        -------
        _Symmetry
            The symmetry of the template configuration.
        """

        return self.configuration.symmetry

    @property
    def system_db(self):
        """The system_db that we belong to."""
        return self._system_db

    @property
    def volume(self):
        """The volume of this template.

        Raises
        ------
        TypeError
            If not a full template or not periodic.

        Returns
        -------
        float
            The volume of the template configuration.
        """

        return self.configuration.volume



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for properties."""

import pprint
import pytest  # noqa: F401


def test_empty(AceticAcid):
    """Test empty properties db."""
    assert AceticAcid.properties.known_properties() == []


def test_add(AceticAcid):
    """Test adding properties db."""
    properties = AceticAcid.properties
    assert not properties.exists("dipole moment")
    properties.add("dipole moment")
    assert properties.known_properties() == ["dipole moment"]
    assert properties.exists("dipole moment")


def test_put(AceticAcid):
    """Test putting some properties and values in."""
    properties = AceticAcid.properties
    properties.add("dipole moment")
    properties.put("dipole moment", 2.0)
    result = properties.get("dipole moment")
    assert result == 2.0


def test_list(AceticAcid):
    """Test a simple query"""
    properties = AceticAcid.properties
    properties.add("dipole moment")
    properties.put("dipole moment", 2.0)

    assert properties.list() == ["dipole moment"]


def test_get_all(AceticAcid):
    """Test getting all the properties and values."""
    answer = {
        "canonical SMILES": "CC(=O)O",
        "dipole moment": 2.0,
        "number of atoms": 8,
    }
    properties = AceticAcid.properties
    properties.add("dipole moment")
    properties.put("dipole moment", 2.0)

    properties.add(
        "number of atoms", "int", description="The number of atoms in the system"
    )
    properties.put("number of atoms", AceticAcid.n_atoms)

    properties.add(
        "canonical SMILES", "str", description="The canonical SMILEs for the structure"
    )
    properties.put("canonical SMILES", AceticAcid.canonical_smiles)

    result = properties.get()
    if result != answer:
        pprint.pprint(result)
    assert result == answer



# -*- coding: utf-8 -*-

"""Property methods for configurations."""

import logging
import pprint  # noqa: F401

logger = logging.getLogger(__name__)


class _ConfigurationProperties(object):
    """A class for handling the properties of a configuration in the star schema."""

    def __init__(self, configuration):
        self._configuration = configuration
        self._cid = configuration.id
        self._properties = None

    @property
    def configuration(self):
        """The configuration these properties are connected with."""
        return self._configuration

    @property
    def properties(self):
        """The general properties object."""
        if self._properties is None:
            self._properties = self.configuration.system_db.properties
        return self._properties

    @property
    def standard_properties(self):
        """A thin wrapper of the _Properties method."""
        return self.properties.standard_properties()

    def add(self, name, _type="float", units=None, description="", noerror=False):
        """A thin wrapper of the _Properties method."""
        return self.properties.add(
            name, _type=_type, units=units, description=description, noerror=noerror
        )

    def exists(self, name):
        """A thin wrapper of the _Properties method."""
        return self.properties.exists(name)

    def get(self, _property="all"):
        """Get the given property value for this configuration.

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        int, float, or str
            The value of the property.
        """
        if _property == "all":
            sql = "SELECT name, type, value"
            sql += "  FROM property, float_data"
            sql += " WHERE float_data.property = property.id and property.id IN ("
            sql += "     SELECT property FROM float_data WHERE configuration = ?"
            sql += ") UNION "
            sql += "SELECT name, type, value"
            sql += "  FROM property, int_data"
            sql += " WHERE int_data.property = property.id and property.id IN ("
            sql += "     SELECT property FROM int_data WHERE configuration = ?"
            sql += ") UNION "
            sql += "SELECT name, type, value"
            sql += "  FROM property, str_data"
            sql += " WHERE str_data.property = property.id and property.id IN ("
            sql += "     SELECT property FROM str_data WHERE configuration = ?"
            sql += ")"

            self.properties.cursor.execute(sql, (self._cid, self._cid, self._cid))

            result = {}
            for row in self.properties.cursor:
                name, _type, value = row
                if _type == "float":
                    result[name] = float(value)
                elif _type == "int":
                    result[name] = int(value)
                else:
                    result[name] = value
            return result
        else:
            return self.properties.get(self._cid, _property)

    def known_properties(self):
        """List the known properties."""
        return self.properties.known_properties()

    def list(self):
        """Return all of the property names for the current configuration.
        Returns
        -------
        [str]
            The names of the properties for the configuration.
        """
        sql = "SELECT name FROM property WHERE id IN ("
        sql += "SELECT property FROM float_data WHERE configuration = ?"
        sql += "UNION SELECT property FROM int_data WHERE configuration = ?"
        sql += "UNION SELECT property FROM str_data WHERE configuration = ?"
        sql += ")"

        self.properties.cursor.execute(sql, (self._cid, self._cid, self._cid))
        return [row[0] for row in self.properties.cursor]

    def list_ids(self):
        """Return all of the property ids for the current configuration.
        Returns
        -------
        [int]
            The names of the properties for the configuration.
        """
        sql = "SELECT property FROM float_data WHERE configuration = ?"
        sql += "UNION SELECT property FROM int_data WHERE configuration = ?"
        sql += "UNION SELECT property FROM str_data WHERE configuration = ?"

        self.properties.cursor.execute(sql, (self._cid, self._cid, self._cid))
        return [row[0] for row in self.properties.cursor]

    def property_id(self, name):
        """The id for a property

        Parameters
        ----------
        name : str
            The name of the property.

        Returns
        -------
        int
            The database id for the property.
        """
        return self.properties.property_id(name)

    def property_name(self, pid):
        """The name of a property

        Parameters
        ----------
        pid : int
            The id of the property.

        Returns
        -------
        str
            The name of the property.
        """
        return self.properties.property_name(pid)

    def property_type(self, _property):
        """The type of a property

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        str
            The type of the property.
        """
        return self.properties.property_type(_property)

    def put(self, _property, value):
        """Store the given property value for this configuration.

        Parameters
        ----------
        _property : int or str
            The id or name of the property.
        value : int, float, or str
            The value to store.
        """
        self.properties.put(self._cid, _property, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the OpenBabel mixin of the class."""

import pprint  # noqa: F401
from openbabel import openbabel
import platform

import pytest  # noqa: F401

# H-Asp-Arg-Val-Tyr-Ile-His-Pro-Phe-OH
SMILES = (
    "[NH2][C@@H](CC(=O)O)C(=O)"
    "[NH][C@@H](CCCNC(=[NH2])N)C(=O)"
    "[NH][C@@H](C(C)C)C(=O)"
    "[NH][C@@H](Cc1ccc(cc1)O)C(=O)"
    "[NH][C@@H]([C@@H](CC)C)C(=O)"
    "[NH][C@@H](CC1=CNC=[NH]1)C(=O)"
    "[N]1[C@@H](CCC1)C(=O)"
    "[NH][C@@H](Cc1ccccc1)C(=O)O"
)
sidechains = {
    "ALA_LL": "[CH3X4]",
    "ARG_LL": "[CH2X4][CH2X4][CH2X4][NH1X3][CH0X3]([NH2X3])=[NH2X3]",
    "ARG_LL_DHH12": "[CH2X4][CH2X4][CH2X4][NH1X3][CH0X3](=[NH1X2])[NH2X3]",
    "ARG_LL_DHH22": "[CH2X4][CH2X4][CH2X4][NH1X3][CH0X3]([NH2X3])=[NH1X2]",
    "ARG_LL_RNH1": "[CH2X4][CH2X4][CH2X4][NH1X3][CH0X3](=[NH2X3])[NH2X3]",
    "ASN_LL": "[CH2X4][CH0X3](=[OH0X1])[NH2X3]",
    "ASP_LL": "[CH2X4][CH0X3](=[OH0X1])[OH1X2]",
    "ASP_LL_DHD2": "[CH2X4][CH0X3](=[OH0X1])[OH0X1]",
    "CYS_LL": "[CH2X4][SH1X2]",
    "CYS_LL_DHG": "[CH2X4][SH0X1]",
    "GLN_LL": "[CH2X4][CH2X4][CH0X3](=[OH0X1])[NH2X3]",
    "GLU_LL": "[CH2X4][CH2X4][CH0X3](=[OH0X1])[OH1X2]",
    "GLU_LL_DHE2": "[CH2X4][CH2X4][CH0X3](=[OH0X1])[OH0X1]",
    "HIS_LL": "[CH2X4][CH0X3][NH1X3]=[CH1X3][NH1X3][CH1X3]",
    "HIS_LL_DHD1": "[CH2X4][CH0X3][NH0X2]=[CH1X3][NH1X3][CH1X3]",
    "HIS_LL_DHE2": "[CH2X4][CH0X3][NH1X3]=[CH1X3][NH0X2][CH1X3]",
    "ILE_LL": "[CH1X4]([CH2X4][CH3X4])[CH3X4]",
    "LEU_LL": "[CH2X4][CH1X4]([CH3X4])[CH3X4]",
    "LYS_LL": "[CH2X4][CH2X4][CH2X4][CH2X4][NH3X4]",
    "LYS_LL_DHZ3": "[CH2X4][CH2X4][CH2X4][CH2X4][NH2X3]",
    "MET_LL": "[CH2X4][CH2X4][SH0X2][CH3X4]",
    "PHE_LL": "[CH2X4][cH0X3][cH1X3][cH1X3][cH1X3][cH1X3][cH1X3]",
    "SER_LL": "[CH2X4][OH1X2]",
    "SER_LL_DHG": "[CH2X4][OH0X1]",
    "THR_LL": "[CH1X4]([OH1X2])[CH3X4]",
    "THR_LL_DHG1": "[CH1X4]([OH0X1])[CH3X4]",
    "TRP_LL": (
        "[CH2X4][CH0X3]=[CH1X3][NH1X3][CH0X3]=[CH0X3][CH1X3]=[CH1X3]" "[CH1X3]=[CH1X3]"
    ),
    "TRP_LL_DHE1": (
        "[CH2X4][CH0X3]=[CH1X3][NH0X2][CH0X3]=[CH0X3][CH1X3]=[CH1X3]" "[CH1X3]=[CH1X3]"
    ),
    "TYR_LL": "[CH2X4][cH0X3]1[cH1X3][cH1X3][cH0X3]([OH1X2])[cH1X3][cH1X3]1",
    "TYR_LL_DHH": ("[CH2X4][CH0X3]=[CH1X3][CH1X3]=[CH0X3]([CH1X3]=[CH1X3])[OH0X1]"),
    "VAL_LL": "[CH1X4]([CH3X4])[CH3X4]",
}

full_smarts = {
    "GLY_LL": "[NH1X2][CH2X4][CH0X2]=[OH0X1]",
    "PRO_LL": "[NH0X2][CH1X4]([CH0X2]=[OH0X1])[CH2X4][CH2X4][CH2X4]",
}

known_input_formats = {}
known_input_formats["Darwin"] = (
    "abinit -- ABINIT Output Format",
    "acesout -- ACES output format",
    "acr -- ACR format",
    "adfband -- ADF Band output format",
    "adfdftb -- ADF DFTB output format",
    "adfout -- ADF output format",
    "alc -- Alchemy format",
    "aoforce -- Turbomole AOFORCE output format",
    "arc -- Accelrys/MSI Biosym/Insight II CAR format",
    "axsf -- XCrySDen Structure Format",
    "bgf -- MSI BGF format",
    "box -- Dock 3.5 Box format",
    "bs -- Ball and Stick format",
    "c09out -- Crystal 09 output format",
    "c3d1 -- Chem3D Cartesian 1 format",
    "c3d2 -- Chem3D Cartesian 2 format",
    "caccrt -- Cacao Cartesian format",
    "can -- Canonical SMILES format",
    "car -- Accelrys/MSI Biosym/Insight II CAR format",
    "castep -- CASTEP format",
    "ccc -- CCC format",
    "cdjson -- ChemDoodle JSON",
    "cdx -- ChemDraw binary format",
    "cdxml -- ChemDraw CDXML format",
    "cif -- Crystallographic Information File",
    "ck -- ChemKin format",
    "cml -- Chemical Markup Language",
    "cmlr -- CML Reaction format",
    "cof -- Culgi object file format",
    "CONFIG -- DL-POLY CONFIG",
    "CONTCAR -- VASP format",
    "CONTFF -- MDFF format",
    "crk2d -- Chemical Resource Kit diagram(2D)",
    "crk3d -- Chemical Resource Kit 3D format",
    "ct -- ChemDraw Connection Table format",
    "cub -- Gaussian cube format",
    "cube -- Gaussian cube format",
    "dallog -- DALTON output format",
    "dalmol -- DALTON input format",
    "dat -- Generic Output file format",
    "dmol -- DMol3 coordinates format",
    "dx -- OpenDX cube format for APBS",
    "ent -- Protein Data Bank format",
    "exyz -- Extended XYZ cartesian coordinates format",
    "fa -- FASTA format",
    "fasta -- FASTA format",
    "fch -- Gaussian formatted checkpoint file format",
    "fchk -- Gaussian formatted checkpoint file format",
    "fck -- Gaussian formatted checkpoint file format",
    "feat -- Feature format",
    "fhiaims -- FHIaims XYZ format",
    "fract -- Free Form Fractional format",
    "fs -- Fastsearch format",
    "fsa -- FASTA format",
    "g03 -- Gaussian Output",
    "g09 -- Gaussian Output",
    "g16 -- Gaussian Output",
    "g92 -- Gaussian Output",
    "g94 -- Gaussian Output",
    "g98 -- Gaussian Output",
    "gal -- Gaussian Output",
    "gam -- GAMESS Output",
    "gamess -- GAMESS Output",
    "gamin -- GAMESS Input",
    "gamout -- GAMESS Output",
    "got -- GULP format",
    "gpr -- Ghemical format",
    "gro -- GRO format",
    "gukin -- GAMESS-UK Input",
    "gukout -- GAMESS-UK Output",
    "gzmat -- Gaussian Z-Matrix Input",
    "hin -- HyperChem HIN format",
    "HISTORY -- DL-POLY HISTORY",
    "inchi -- InChI format",
    "inp -- GAMESS Input",
    "ins -- ShelX format",
    "jin -- Jaguar input format",
    "jout -- Jaguar output format",
    "log -- Generic Output file format",
    "lpmd -- LPMD format",
    "mcdl -- MCDL format",
    "mcif -- Macromolecular Crystallographic Info",
    "MDFF -- MDFF format",
    "mdl -- MDL MOL format",
    "ml2 -- Sybyl Mol2 format",
    "mmcif -- Macromolecular Crystallographic Info",
    "mmd -- MacroModel format",
    "mmod -- MacroModel format",
    "mol -- MDL MOL format",
    "mol2 -- Sybyl Mol2 format",
    "mold -- Molden format",
    "molden -- Molden format",
    "molf -- Molden format",
    "moo -- MOPAC Output format",
    "mop -- MOPAC Cartesian format",
    "mopcrt -- MOPAC Cartesian format",
    "mopin -- MOPAC Internal",
    "mopout -- MOPAC Output format",
    "mpc -- MOPAC Cartesian format",
    "mpo -- Molpro output format",
    "mpqc -- MPQC output format",
    "mrv -- Chemical Markup Language",
    "msi -- Accelrys/MSI Cerius II MSI format",
    "nwo -- NWChem output format",
    "orca -- ORCA output format",
    "out -- Generic Output file format",
    "outmol -- DMol3 coordinates format",
    "output -- Generic Output file format",
    "pc -- PubChem format",
    "pcjson -- PubChem JSON",
    "pcm -- PCModel Format",
    "pdb -- Protein Data Bank format",
    "pdbqt -- AutoDock PDBQT format",
    "png -- PNG 2D depiction",
    "pos -- POS cartesian coordinates format",
    "POSCAR -- VASP format",
    "POSFF -- MDFF format",
    "pqr -- PQR format",
    "pqs -- Parallel Quantum Solutions format",
    "prep -- Amber Prep format",
    "pwscf -- PWscf format",
    "qcout -- Q-Chem output format",
    "res -- ShelX format",
    "rsmi -- Reaction SMILES format",
    "rxn -- MDL RXN format",
    "sd -- MDL MOL format",
    "sdf -- MDL MOL format",
    "siesta -- SIESTA format",
    "smi -- SMILES format",
    "smiles -- SMILES format",
    "smy -- SMILES format using Smiley parser",
    "sy2 -- Sybyl Mol2 format",
    "t41 -- ADF TAPE41 format",
    "tdd -- Thermo format",
    "text -- Read and write raw text",
    "therm -- Thermo format",
    "tmol -- TurboMole Coordinate format",
    "txt -- Title format",
    "txyz -- Tinker XYZ format",
    "unixyz -- UniChem XYZ format",
    "VASP -- VASP format",
    "vmol -- ViewMol format",
    "wln -- Wiswesser Line Notation",
    "xml -- General XML format",
    "xsf -- XCrySDen Structure Format",
    "xyz -- XYZ cartesian coordinates format",
    "yob -- YASARA.org YOB format",
)
known_input_formats["Linux"] = (
    "abinit -- ABINIT Output Format",
    "acesout -- ACES output format",
    "acr -- ACR format",
    "adfband -- ADF Band output format",
    "adfdftb -- ADF DFTB output format",
    "adfout -- ADF output format",
    "alc -- Alchemy format",
    "aoforce -- Turbomole AOFORCE output format",
    "arc -- Accelrys/MSI Biosym/Insight II CAR format",
    "axsf -- XCrySDen Structure Format",
    "bgf -- MSI BGF format",
    "box -- Dock 3.5 Box format",
    "bs -- Ball and Stick format",
    "c09out -- Crystal 09 output format",
    "c3d1 -- Chem3D Cartesian 1 format",
    "c3d2 -- Chem3D Cartesian 2 format",
    "caccrt -- Cacao Cartesian format",
    "can -- Canonical SMILES format",
    "car -- Accelrys/MSI Biosym/Insight II CAR format",
    "castep -- CASTEP format",
    "ccc -- CCC format",
    "cdjson -- ChemDoodle JSON",
    "cdx -- ChemDraw binary format",
    "cdxml -- ChemDraw CDXML format",
    "cif -- Crystallographic Information File",
    "ck -- ChemKin format",
    "cml -- Chemical Markup Language",
    "cmlr -- CML Reaction format",
    "cof -- Culgi object file format",
    "CONFIG -- DL-POLY CONFIG",
    "CONTCAR -- VASP format",
    "CONTFF -- MDFF format",
    "crk2d -- Chemical Resource Kit diagram(2D)",
    "crk3d -- Chemical Resource Kit 3D format",
    "ct -- ChemDraw Connection Table format",
    "cub -- Gaussian cube format",
    "cube -- Gaussian cube format",
    "dallog -- DALTON output format",
    "dalmol -- DALTON input format",
    "dat -- Generic Output file format",
    "dmol -- DMol3 coordinates format",
    "dx -- OpenDX cube format for APBS",
    "ent -- Protein Data Bank format",
    "exyz -- Extended XYZ cartesian coordinates format",
    "fa -- FASTA format",
    "fasta -- FASTA format",
    "fch -- Gaussian formatted checkpoint file format",
    "fchk -- Gaussian formatted checkpoint file format",
    "fck -- Gaussian formatted checkpoint file format",
    "feat -- Feature format",
    "fhiaims -- FHIaims XYZ format",
    "fract -- Free Form Fractional format",
    "fs -- Fastsearch format",
    "fsa -- FASTA format",
    "g03 -- Gaussian Output",
    "g09 -- Gaussian Output",
    "g16 -- Gaussian Output",
    "g92 -- Gaussian Output",
    "g94 -- Gaussian Output",
    "g98 -- Gaussian Output",
    "gal -- Gaussian Output",
    "gam -- GAMESS Output",
    "gamess -- GAMESS Output",
    "gamin -- GAMESS Input",
    "gamout -- GAMESS Output",
    "got -- GULP format",
    "gpr -- Ghemical format",
    "gro -- GRO format",
    "gukin -- GAMESS-UK Input",
    "gukout -- GAMESS-UK Output",
    "gzmat -- Gaussian Z-Matrix Input",
    "hin -- HyperChem HIN format",
    "HISTORY -- DL-POLY HISTORY",
    "inchi -- InChI format",
    "inp -- GAMESS Input",
    "ins -- ShelX format",
    "jin -- Jaguar input format",
    "jout -- Jaguar output format",
    "log -- Generic Output file format",
    "lpmd -- LPMD format",
    "mcdl -- MCDL format",
    "mcif -- Macromolecular Crystallographic Info",
    "MDFF -- MDFF format",
    "mdl -- MDL MOL format",
    "ml2 -- Sybyl Mol2 format",
    "mmcif -- Macromolecular Crystallographic Info",
    "mmd -- MacroModel format",
    "mmod -- MacroModel format",
    "mol -- MDL MOL format",
    "mol2 -- Sybyl Mol2 format",
    "mold -- Molden format",
    "molden -- Molden format",
    "molf -- Molden format",
    "moo -- MOPAC Output format",
    "mop -- MOPAC Cartesian format",
    "mopcrt -- MOPAC Cartesian format",
    "mopin -- MOPAC Internal",
    "mopout -- MOPAC Output format",
    "mpc -- MOPAC Cartesian format",
    "mpo -- Molpro output format",
    "mpqc -- MPQC output format",
    "mrv -- Chemical Markup Language",
    "msi -- Accelrys/MSI Cerius II MSI format",
    "nwo -- NWChem output format",
    "orca -- ORCA output format",
    "out -- Generic Output file format",
    "outmol -- DMol3 coordinates format",
    "output -- Generic Output file format",
    "pc -- PubChem format",
    "pcjson -- PubChem JSON",
    "pcm -- PCModel Format",
    "pdb -- Protein Data Bank format",
    "pdbqt -- AutoDock PDBQT format",
    "png -- PNG 2D depiction",
    "pos -- POS cartesian coordinates format",
    "POSCAR -- VASP format",
    "POSFF -- MDFF format",
    "pqr -- PQR format",
    "pqs -- Parallel Quantum Solutions format",
    "prep -- Amber Prep format",
    "pwscf -- PWscf format",
    "qcout -- Q-Chem output format",
    "res -- ShelX format",
    "rsmi -- Reaction SMILES format",
    "rxn -- MDL RXN format",
    "sd -- MDL MOL format",
    "sdf -- MDL MOL format",
    "siesta -- SIESTA format",
    "smi -- SMILES format",
    "smiles -- SMILES format",
    "smy -- SMILES format using Smiley parser",
    "sy2 -- Sybyl Mol2 format",
    "t41 -- ADF TAPE41 format",
    "tdd -- Thermo format",
    "text -- Read and write raw text",
    "therm -- Thermo format",
    "tmol -- TurboMole Coordinate format",
    "txt -- Title format",
    "txyz -- Tinker XYZ format",
    "unixyz -- UniChem XYZ format",
    "VASP -- VASP format",
    "vmol -- ViewMol format",
    "wln -- Wiswesser Line Notation",
    "xml -- General XML format",
    "xsf -- XCrySDen Structure Format",
    "xtc -- XTC format",
    "xyz -- XYZ cartesian coordinates format",
    "yob -- YASARA.org YOB format",
)
known_input_formats["Windows"] = known_input_formats["Darwin"]


def test_substructure(CH3COOH_3H2O):
    """Test the finding substructures in the configuration."""
    answer1 = [(5, 6, 7)]
    answer2 = [(5, 6, 7, 8)]
    answer3 = [(9,), (12,), (15,)]
    answer4 = [(10, 9, 11), (13, 12, 14), (16, 15, 17)]
    answer5 = [(9, 10, 11), (12, 13, 14), (15, 16, 17)]

    configuration = CH3COOH_3H2O

    # Just the C and O of the carboxyl, not to H on O
    result = configuration.find_substructures("C(=O)O")
    if result != answer1:
        pprint.pprint(result)
    assert result == answer1

    # All four atoms of the carboxyl group
    result = configuration.find_substructures("C(=O)[O][H]")
    if result != answer2:
        pprint.pprint(result)
    assert result == answer2

    # The Oxygen atoms of the waters
    result = configuration.find_substructures("[OH2]")
    if result != answer3:
        pprint.pprint(result)
    assert result == answer3

    # All the atoms in waters, order H-O-H
    result = configuration.find_substructures("[H][O][H]")
    if result != answer4:
        pprint.pprint(result)
    assert result == answer4

    # All the atoms in waters, order O-H-H
    result = configuration.find_substructures("[O]([H])[H]")
    if result != answer5:
        pprint.pprint(result)
    assert result == answer5


def test_substructure_ordering(disordered):
    """Test the ordering of atoms in subsets."""
    answer1 = "CC(=O)O"
    answer2 = [(1, 5, 6, 7), (16, 12, 11, 10)]
    answer3 = "C([H])([H])([H])C(=O)O[H]"
    answer4 = [(1, 2, 3, 4, 5, 6, 7, 8), (16, 15, 14, 13, 12, 11, 10, 9)]

    configuration = disordered
    templates = configuration.create_molecule_templates(create_subsets=False)

    # Without hydrogens
    smiles = templates[0].smiles
    if smiles != answer1:
        pprint.pprint(smiles)
    assert smiles == answer1

    result = configuration.find_substructures(smiles)
    if result != answer2:
        pprint.pprint(result)
    assert result == answer2

    # With hydrogens
    smiles = templates[0].to_smiles(hydrogens=True)
    if smiles != answer3:
        pprint.pprint(smiles)
    assert smiles == answer3

    result = configuration.find_substructures(smiles)
    if result != answer4:
        pprint.pprint(result)
    assert result == answer4


def test_all_residue_search(configuration):
    """Testing locating residues in a peptide."""
    residues = {
        "ARG_LL": [(9, 10, 11, 12, 13, 14, 15, 17, 16, 18, 19)],
        "ARG_LL_RNH1": [(9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19)],
        "HIS_LL": [(47, 48, 49, 50, 54, 53, 52, 51, 55, 56)],
        "ILE_LL": [(39, 40, 41, 42, 43, 44, 45, 46)],
        "PHE_LL": [(64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74)],
        "TYR_LL": [(27, 28, 29, 30, 31, 32, 33, 36, 34, 35, 37, 38)],
        "VAL_LL": [(20, 21, 22, 23, 24, 25, 26)],
    }
    n_terminal = {
        "ASP_LL": [(1, 2, 3, 4, 5, 6, 7, 8)],
    }
    c_terminal = {
        "PHE_LL": [(64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75)],
    }

    configuration.from_smiles(SMILES)

    for name, sc in sidechains.items():
        smarts = f"[NH1X3][C@@H]({sc})[CX3]=[OX1]"
        result = configuration.find_substructures(smarts)
        if len(result) == 0:
            assert name not in residues
        else:
            if result != residues[name]:
                print(f"'{name}': {result},")
            assert result == residues[name]

    for name, sc in sidechains.items():
        smarts = f"[NH2][C@@H]({sc})[CX3]=[OX1]"
        result = configuration.find_substructures(smarts)
        if len(result) == 0:
            assert name not in n_terminal
        else:
            if result != n_terminal[name]:
                print(f"'{name}': {result},")
            assert result == n_terminal[name]

    for name, sc in sidechains.items():
        smarts = f"[NX3][C@@H]({sc})[CX3](=[OX1])[OH1X2]"
        result = configuration.find_substructures(smarts)
        if len(result) == 0:
            assert name not in c_terminal
        else:
            if result != c_terminal[name]:
                print(f"'{name}': {result},")
            assert result == c_terminal[name]


def test_to_OBMol(configuration):
    """Test creating an OBMol object from a structure."""
    mol = configuration.to_OBMol()

    bondorder_list = []
    for bond in openbabel.OBMolBondIter(mol):
        bondorder_list.append(bond.GetBondOrder())

    atno_list = []
    for atno in openbabel.OBMolAtomIter(mol):
        atno_list.append(mol.GetAtmoicNum())

    assert configuration.atoms.atomic_numbers == atno_list
    assert configuration.bonds.bondorders == bondorder_list
    # assert configuration.n_atoms == mol.NumAtoms()
    # assert configuration.n_bonds == mol.NumBonds()


def test_from_OBMol(configuration):
    """Test creating a structure from an OBMol object."""
    obConversion = openbabel.OBConversion()
    obConversion.SetInAndOutFormats("smi", "mdl")
    obConversion.AddOption("3")
    mol = openbabel.OBMol()
    obConversion.ReadString(mol, "C=CO")

    # Add hydrogens
    mol.AddHydrogens()

    # Get coordinates for a 3-D structure
    builder = openbabel.OBBuilder()
    builder.Build(mol)

    configuration.from_OBMol(mol)

    assert configuration.n_atoms == 7
    assert configuration.bonds.bondorders == [2, 1, 1, 1, 1, 1]


def test_input_formats():
    """Check the list of input formats Open Babel handles"""
    obConversion = openbabel.OBConversion()
    formats = obConversion.GetSupportedInputFormat()
    system = platform.system()
    if formats != known_input_formats[system]:
        import pprint

        print(system)
        pprint.pprint(formats)
    assert formats == known_input_formats[system]



# -*- coding: utf-8 -*-

"""Functions for handling PDB files

To Do
^^^^^

Need to understand more fully the PDB/mmcif format and the how to
carry the information about residues, chains, hetero groups, waters, etc.
At the moment this is ignoring much of the information, and putting residue,
chain, etc information directly on atoms.

I think we should use templates and subsets, but am not (yet) sure.

Presumably this metadata is most useful for setting up complicated
simulations.

File Format
^^^^^^^^^^^

For complete documentation, see
http://www.wwpdb.org/documentation/file-format-content/format33/v3.3.html

Order of records::

    RECORD TYPE             EXISTENCE           CONDITIONS IF  OPTIONAL
    --------------------------------------------------------------------------------------
    HEADER                  Mandatory
    OBSLTE                  Optional            Mandatory in  entries that have been
                                                replaced by a newer entry.
    TITLE                   Mandatory
    SPLIT                   Optional            Mandatory when  large macromolecular
                                                complexes  are split into multiple PDB
                                                entries.
    CAVEAT                  Optional            Mandatory when there are outstanding  errors
                                                such  as chirality.
    COMPND                  Mandatory
    SOURCE                  Mandatory
    KEYWDS                  Mandatory
    EXPDTA                  Mandatory
    NUMMDL                  Optional            Mandatory for  NMR ensemble entries.
    MDLTYP                  Optional            Mandatory for  NMR minimized average
                                                Structures or when the entire  polymer
                                                chain contains C alpha or P atoms only.
    AUTHOR                  Mandatory
    REVDAT                  Mandatory
    SPRSDE                  Optional            Mandatory for a replacement entry.
    JRNL                    Optional            Mandatory for a publication describes
                                                the experiment.
    REMARK 0                Optional            Mandatory for a re-refined structure
    REMARK 1                Optional
    REMARK 2                Mandatory
    REMARK 3                Mandatory
    REMARK N                Optional            Mandatory under certain conditions.
    DBREF                   Optional            Mandatory for all polymers.
    DBREF1/DBREF2           Optional            Mandatory when certain sequence  database
                                                accession  and/or sequence numbering
                                                does  not fit preceding DBREF format.
    SEQADV                  Optional            Mandatory if sequence  conflict exists.
    SEQRES                  Mandatory           Mandatory if ATOM records exist.
    MODRES                  Optional            Mandatory if modified group exists  in the
                                                coordinates.
    HET                     Optional            Mandatory if a non-standard group other
                                                than water appears in the coordinates.
    HETNAM                  Optional            Mandatory if a non-standard group other
                                                than  water appears in the coordinates.
    HETSYN                  Optional
    FORMUL                  Optional            Mandatory if a non-standard group or
                                                water appears in the coordinates.
    HELIX                   Optional
    SHEET                   Optional
    SSBOND                  Optional            Mandatory if a  disulfide bond is present.
    LINK                    Optional            Mandatory if  non-standard residues appear
                                                in a  polymer
    CISPEP                  Optional
    SITE                    Optional
    CRYST1                  Mandatory
    ORIGX1 ORIGX2 ORIGX3    Mandatory
    SCALE1 SCALE2 SCALE3    Mandatory
    MTRIX1 MTRIX2 MTRIX3    Optional            Mandatory if  the complete asymmetric unit
                                                must  be generated from the given coordinates
                                                using non-crystallographic symmetry.
    MODEL                   Optional            Mandatory if more than one model
                                                is  present in the entry.
    ATOM                    Optional            Mandatory if standard residues exist.
    ANISOU                  Optional
    TER                     Optional            Mandatory if ATOM records exist.
    HETATM                  Optional            Mandatory if non-standard group exists.
    ENDMDL                  Optional            Mandatory if MODEL appears.
    CONECT                  Optional            Mandatory if non-standard group appears
                                                and  if LINK or SSBOND records exist.
    MASTER                  Mandatory
    END                     Mandatory

Description of HETATM records::

    COLUMNS       DATA  TYPE     FIELD         DEFINITION
    -----------------------------------------------------------------------
     1 - 6        Record name    "HETATM"
     7 - 11       Integer        serial        Atom serial number.
    13 - 16       Atom           name          Atom name.
    17            Character      altLoc        Alternate location indicator.
    18 - 20       Residue name   resName       Residue name.
    22            Character      chainID       Chain identifier.
    23 - 26       Integer        resSeq        Residue sequence number.
    27            AChar          iCode         Code for insertion of residues.
    31 - 38       Real(8.3)      x             Orthogonal coordinates for X.
    39 - 46       Real(8.3)      y             Orthogonal coordinates for Y.
    47 - 54       Real(8.3)      z             Orthogonal coordinates for Z.
    55 - 60       Real(6.2)      occupancy     Occupancy.
    61 - 66       Real(6.2)      tempFactor    Temperature factor.
    77 - 78       LString(2)     element       Element symbol; right-justified.
    79 - 80       LString(2)     charge        Charge on the atom.

Description of CONECT records::

    COLUMNS       DATA  TYPE      FIELD        DEFINITION
    -------------------------------------------------------------------------
     1 -  6        Record name    "CONECT"
     7 - 11        Integer        serial       Atom  serial number
    12 - 16        Integer        serial       Serial number of bonded atom
    17 - 21        Integer        serial       Serial  number of bonded atom
    22 - 26        Integer        serial       Serial number of bonded atom
    27 - 31        Integer        serial       Serial number of bonded atom
"""  # noqa: E501

import collections
import logging
import time

logger = logging.getLogger(__name__)


class PDBMixin:
    """A mixin for handling PDB files."""

    def to_pdb_text(self, title=None, comment="Exported from SEAMM"):
        """Create the text of the PDB file from the system.

        Parameters
        ----------
        title : str = None
            The title for the structure, by default the system name.
        comment : str = 'Exported from SEAMM'
            Comment line

        Returns
        -------
        text : str
            The text of the file.
        """
        lines = []

        atoms = self.atoms

        n_atoms = atoms.n_atoms

        date_time = time.strftime("%m%d%y%H%M")

        lines.append("COMPND    UNNAMED")
        lines.append("AUTHOR    MolSSI SEAMM at " + date_time)

        # atoms

        if "resname" in atoms:
            resnames = atoms["resname"]
        else:
            resnames = ["UNK"] * n_atoms

        if "chainid" in atoms:
            chainids = atoms["chainid"]
        else:
            chainids = ["A"] * n_atoms

        if "resseq" in atoms:
            resseqs = [1 if x is None else x for x in atoms["resseq"]]
        else:
            resseqs = [1] * n_atoms

        if "occupancy" in atoms:
            occupancies = atoms["occupancy"]
        else:
            occupancies = [1.0] * n_atoms

        if "tempfactor" in atoms:
            tempfactors = atoms["tempfactor"]
        else:
            tempfactors = [0.0] * n_atoms

        if "formal_charge" in atoms:
            charges = atoms["formal_charge"]
        else:
            charges = [" "] * n_atoms

        count = 0
        symbols = atoms.symbols
        coordinates = atoms.coordinates
        if "name" in atoms:
            names = [
                symbol if name is None else name
                for name, symbol in zip(atoms["name"], symbols)
            ]
        else:
            names = symbols

        for (
            element,
            xyz,
            name,
            resname,
            chainid,
            resseq,
            occupancy,
            tempfactor,
            charge,
        ) in zip(
            symbols,
            coordinates,
            names,
            resnames,
            chainids,
            resseqs,
            occupancies,
            tempfactors,
            charges,
        ):
            count += 1
            x, y, z = xyz
            if len(element) == 1 and len(name) < 4:
                name = " " + name
            lines.append(
                f"ATOM  {count:5d} {name:4s} {resname:3s} {chainid:1s}"
                f"{resseq:4d}    {x:8.3f}{y:8.3f}{z:8.3f}{occupancy:6.2f}"
                f"{tempfactor:6.2f}          {element.upper():>2s}"
                f"{charge:2}"
            )

        # bonds
        for i, js in enumerate(self.bonded_neighbors(as_indices=True)):
            if len(js) > 0:
                lines.append(f"CONECT{i+1:5d}" + "".join([f"{j+1:5d}" for j in js]))

        lines.append(
            "MASTER        0    0    0    0    0    0    0    0"
            f"{n_atoms:5d}    0{n_atoms:5d}"
        )
        lines.append("END")

        return "\n".join(lines)

    def from_pdb_text(self, data):
        """Create the system from a PDF file.

        Parameters
        ----------
        data : str
            The complete text of the Molfile.
        """
        # Initialize the structure

        self.clear()
        self.periodicity = 0

        if isinstance(data, list):
            lines = data
        else:
            lines = data.splitlines()

        names = []
        symbols = []
        xs = []
        ys = []
        zs = []
        resnames = []
        chainids = []
        resseqs = []
        occupancies = []
        tempfactors = []
        connections = None
        for line in lines:
            key = line[0:6].rstrip()
            if key == "HEADER":
                pass
            elif key == "OBSLTE":
                pass
            elif key == "TITLE":
                pass
            elif key == "SPLIT":
                pass
            elif key == "CAVEAT":
                pass
            elif key == "COMPND":
                pass
            elif key == "SOURCE":
                pass
            elif key == "KEYWDS":
                pass
            elif key == "EXPDTA":
                pass
            elif key == "NUMMDL":
                pass
            elif key == "MDLTYPE":
                pass
            elif key == "AUTHOR":
                pass
            elif key == "REVDAT":
                pass
            elif key == "SPRSDE":
                pass
            elif key == "JRNL":
                pass
            elif key == "REMARK":
                pass
            elif key == "DBREF":
                pass
            elif key == "DBREF1":
                pass
            elif key == "DBREF2":
                pass
            elif key == "SEQADV":
                pass
            elif key == "SEQRES":
                pass
            elif key == "MODRES":
                pass
            elif key == "HET":
                pass
            elif key == "HETNAM":
                pass
            elif key == "HETSYN":
                pass
            elif key == "FORMUL":
                pass
            elif key == "HELIX":
                pass
            elif key == "SHEET":
                pass
            elif key == "SSBOND":
                pass
            elif key == "LINK":
                pass
            elif key == "CISPEP":
                pass
            elif key == "SITE":
                pass
            elif key == "CRYST1":
                pass
            elif key == "ORIGX1":
                pass
            elif key == "ORIGX2":
                pass
            elif key == "ORIGX3":
                pass
            elif key == "SCALE1":
                pass
            elif key == "SCALE2":
                pass
            elif key == "SCALE3":
                pass
            elif key == "MTRIX1":
                pass
            elif key == "MTRIX2":
                pass
            elif key == "MTRIX3":
                pass
            elif key == "MODEL":
                pass
            elif key == "ATOM" or key == "HETATM":
                # serial = int(line[6:11])  # noqa: F841
                name = line[12:16].strip()
                # altloc = line[16]  # noqa: F841
                resname = line[17:20].strip()
                chainid = line[21]
                resseq = int(line[22:26])
                # icode = line[26]  # noqa: F841
                x = float(line[30:38])
                y = float(line[38:46])
                z = float(line[46:54])
                tmp = line[54:60].strip()
                occupancy = 1.0 if tmp == "" else float(tmp)
                tmp = line[60:66].strip()
                tempfactor = 0.0 if tmp == "" else float(tmp)

                # Symbol maybe fully capitalized e.g. 'FE', so need to fix
                symbol = line[75:78].strip().capitalize()
                # tmp = line[78:80].strip()
                # charge = 0.0 if tmp == '' else float(tmp)  # noqa: F841

                names.append(name)
                symbols.append(symbol)
                xs.append(x)
                ys.append(y)
                zs.append(z)
                resnames.append(resname)
                chainids.append(chainid)
                resseqs.append(resseq)
                occupancies.append(occupancy)
                tempfactors.append(tempfactor)
            elif key == "ANISOU":
                pass
            elif key == "TER":
                pass
            elif key == "ENDMDL":
                pass
            elif key == "CONECT":
                if connections is None:
                    n_atoms = len(symbols)
                    connections = []
                    for i in range(n_atoms + 1):
                        connections.append(list())

                atom = int(line[6:11])
                # should use columns in case they run together.
                for i in range(11, 31, 5):
                    tmp = line[i : i + 5].strip()
                    if tmp == "":
                        break
                    connections[atom].append(int(tmp))
            elif key == "MASTER":
                pass
            elif key == "END":
                break
            else:
                raise RuntimeError("Illegal line in PDB file\n\t" + line)

        if "name" not in self.atoms:
            self.atoms.add_attribute("name", coltype="string")

        atom_id = self.atoms.append(symbol=symbols, name=names, x=xs, y=ys, z=zs)

        if "resname" in self.atoms:
            self.atoms["resname"][0:] = resnames
        else:
            counts = collections.Counter(resnames)
            if len(counts) > 1 or [*counts.keys()] != ["UNK"]:
                self.atoms.add_attribute("resname", coltype="string")
                self.atoms["resname"][0:] = resnames

        if "chainid" in self.atoms:
            self.atoms["chainid"][0:] = chainids
        else:
            counts = collections.Counter(chainids)
            if len(counts) > 1 or [*counts.keys()] != ["A"]:
                self.atoms.add_attribute("chainid", coltype="string")
                self.atoms["chainid"][0:] = chainids

        if "resseq" in self.atoms:
            self.atoms["resseq"][0:] = resseqs
        else:
            counts = collections.Counter(resseqs)
            if len(counts) > 1 or [*counts.keys()] != ["1"]:
                self.atoms.add_attribute("resseq", coltype="int", default=1)
                self.atoms["resseq"][0:] = resseqs

        if "occupancy" in self.atoms:
            self.atoms["occupancy"][0:] = occupancies
        else:
            counts = collections.Counter(occupancies)
            if len(counts) > 1 or [*counts.keys()] != [1.0]:
                self.atoms.add_attribute("occupancy", coltype="float", default=1.0)
                self.atoms["occupancy"][0:] = occupancies

        if "tempfactor" in self.atoms:
            self.atoms["tempfactor"][0:] = tempfactors
        else:
            counts = collections.Counter(tempfactors)
            if len(counts) > 1 or [*counts.keys()] != [0.0]:
                self.atoms.add_attribute("tempfactor", coltype="float", default=1.0)
                self.atoms["tempfactor"][0:] = tempfactors

        iatom = []
        jatom = []
        if connections is not None:
            for i in range(1, n_atoms + 1):
                for j in connections[i]:
                    if i not in connections[j]:
                        logger.warning(f"Bond {j}-{i} not found in PDB file")
                        # put in the bond since we won't see its partner!
                        iatom.append(atom_id[i - 1])
                        jatom.append(atom_id[j - 1])
                    elif i < j:
                        # put in only 1 of 2 equivalent bonds
                        iatom.append(atom_id[i - 1])
                        jatom.append(atom_id[j - 1])

            self.bonds.append(i=iatom, j=jatom)



# -*- coding: utf-8 -*-

"""Functions for handling CIF files

Bond Orders
-----------
1 sing	single bond
2 doub	double bond
3 trip	triple bond
4 quad	quadruple bond
5 arom	aromatic bond
6 delo	delocalized double bond
7 pi	pi bond
8 poly	polymeric bond
"""

import io
import logging
import math

import CifFile

logger = logging.getLogger(__name__)

bond_order = {
    1: "SING",
    2: "DOUB",
    3: "TRIP",
    4: "QUAD",
    5: "AROM",
    6: "DELO",
    7: "PI",
    8: "POLY",
}
to_bond_order = {j: i for i, j in bond_order.items()}


class CIFMixin:
    """A mixin for handling CIF files."""

    def read_cif_file(self, path):
        """Create new systems from a CIF file.

        Read a CIF file and create a new system from each datablock in
        the file.

        If the datablock has an ensemble, as denoted by a section
        '_pdbx_nmr_ensemble', a configuration will be created for each
        conformer. If there is a representative conformer, the current
        configuration will point to it; otherwise to the last conformer.

        Parameters
        ----------
        path : str or Path
            A string or Path object pointing to the file to be read.

        Returns
        -------
        [_System]
            List of systems created.
        """
        if "SystemDB" in str(type(self)):
            sys_db = self
        else:
            sys_db = self.system_db

        lines = []
        systems = []
        configurations = {}
        in_block = False
        block_name = ""
        with open(path, "r") as fd:
            for line in fd:
                if line[0:5] == "data_":
                    self.logger.debug(f"Found block {line}")
                    if not in_block:
                        in_block = True
                    else:
                        new_systems, new_configurations = self.from_mmcif_text(
                            "\n".join(lines)
                        )
                        self.logger.debug(
                            f"   added system {sys_db.n_systems}: {block_name}"
                        )
                        systems.extend(new_systems)
                        configurations.update(new_configurations)
                    block_name = line[5:].strip()
                    lines = []
                lines.append(line)

            if len(lines) > 0:
                # The last block just ends at the end of the file
                new_systems, new_configurations = self.from_mmcif_text("\n".join(lines))
                self.logger.debug(f"   added system {sys_db.n_systems}: {block_name}")
                systems.extend(new_systems)
                configurations.update(new_configurations)

        return systems, configurations

    def to_cif_text(self):
        """Create the text of a CIF file from this configuration.

        Returns
        -------
        text : str
            The text of the file.
        """

        lines = []

        atoms = self.atoms

        # Get the chemical formula
        formula, empirical_formula, Z = self.formula

        _id = empirical_formula.replace(" ", "")

        # And created the file, line-by-line
        lines = []
        lines.append("# Generated by MolSSI SEAMM")
        lines.append(f"data_{_id}")

        # Cell information
        if self.periodicity == 3:
            cell = self.cell
            a, b, c, alpha, beta, gamma = cell.parameters
            volume = cell.volume
            lines.append("_symmetry_space_group_name_H-M   'P 1'")
            lines.append(f"_cell_length_a   {a}")
            lines.append(f"_cell_length_b   {b}")
            lines.append(f"_cell_length_c   {c}")
            lines.append(f"_cell_angle_alpha   {alpha}")
            lines.append(f"_cell_angle_beta    {beta}")
            lines.append(f"_cell_angle_gamma   {gamma}")
            lines.append("_symmetry_Int_Tables_number   1")
            lines.append(f"_cell_volume   {volume}")
            lines.append(f"_cell_formula_units_Z   {Z}")
            lines.append("loop_")
            lines.append(" _symmetry_equiv_pos_site_id")
            lines.append(" _symmetry_equiv_pos_as_xyz")
            lines.append("  1  'x, y, z'")

        lines.append(f"_chemical_formula_structural   {empirical_formula}")
        lines.append(f"_chemical_formula_sum   '{formula}'")

        # The atoms
        lines.append("loop_")
        lines.append(" _atom_site_type_symbol")
        lines.append(" _atom_site_label")
        lines.append(" _atom_site_symmetry_multiplicity")
        lines.append(" _atom_site_fract_x")
        lines.append(" _atom_site_fract_y")
        lines.append(" _atom_site_fract_z")
        lines.append(" _atom_site_occupancy")

        # Need unique names
        if "names" in atoms:
            original_names = atoms.get_column("names")
        else:
            original_names = atoms.symbols

        names = []
        tmp = {}
        for name in original_names:
            if name in tmp:
                tmp[name] += 1
                names.append(name + str(tmp[name]))
            else:
                tmp[name] = 1
                names.append(name)

        UVW = atoms.get_coordinates(fractionals=True, in_cell="molecule")

        symbols = atoms.symbols
        for element, name, uvw in zip(symbols, names, UVW):
            u, v, w = uvw
            lines.append(f"{element} {name}  1  {u:.3f} {v:.3f} {w:.3f}  1")

        # And that is it!
        return "\n".join(lines)

    def from_cif_text(self, text):
        """Create this configuration from a CIF file..

        Parameters
        ----------
        text : str
            The text from the CIF file

        Returns
        -------
        None
        """

        cif = CifFile.ReadCif(io.StringIO(text))

        data_blocks = [*cif.keys()]

        if len(data_blocks) != 1:
            raise RuntimeError(
                f"There are {len(data_blocks)} data blocks in the cif file."
            )
        data_block = cif[data_blocks[0]]

        # Reset the system
        self.clear()
        self.periodicity = 3
        self.coordinate_system = "fractional"

        # The cell
        a = data_block["_cell_length_a"]
        b = data_block["_cell_length_b"]
        c = data_block["_cell_length_c"]
        alpha = data_block["_cell_angle_alpha"]
        beta = data_block["_cell_angle_beta"]
        gamma = data_block["_cell_angle_gamma"]
        self.cell.parameters = (a, b, c, alpha, beta, gamma)

        # Add the atoms
        # TEMPORARILY lower the symmetry to P1
        delta = 1.0e-04

        # Where is the symmetry info?
        if "_space_group_symop_operation_xyz" in data_block:
            symdata = "_space_group_symop_operation_xyz"
        elif "_symmetry_equiv_pos_as_xyz" in data_block:
            symdata = "_symmetry_equiv_pos_as_xyz"
        else:
            raise RuntimeError(
                "CIF file does not contain required symmetry information. Neither "
                "'_symmetry_equiv_pos_as_xyz' or '_space_group_symop_operation_xyz' "
                "is present."
            )
        xs = []
        ys = []
        zs = []
        symbols = []
        # May have type symbols or labels, or both. Use type symbols by preference
        if "_atom_site_type_symbol" in data_block:
            type_section = "_atom_site_type_symbol"
        elif "_atom_site_label" in data_block:
            type_section = "_atom_site_label"
        else:
            raise KeyError(
                "Neither _atom_site_type_label or _atom_site_label are in file"
            )
        for x, y, z, symbol in zip(
            data_block["_atom_site_fract_x"],
            data_block["_atom_site_fract_y"],
            data_block["_atom_site_fract_z"],
            data_block[type_section],
        ):  # yapf: disable
            # Atom symbols may be followed by number, etc.
            if len(symbol) > 1:
                if symbol[1].isalpha():
                    symbol = symbol[0:2]
                else:
                    symbol = symbol[0]
            # These variables *are* used in the eval below.
            x = float(x)
            y = float(y)
            z = float(z)
            for symop in data_block[symdata]:
                x_eq, y_eq, z_eq = symop.split(",")
                x_new = eval(x_eq)
                y_new = eval(y_eq)
                z_new = eval(z_eq)
                # Translate into cell.
                x_new = x_new - math.floor(x_new)
                y_new = y_new - math.floor(y_new)
                z_new = z_new - math.floor(z_new)
                # check for almost 1, should be 0
                if abs(1 - x_new) < delta:
                    x_new = 0.0
                if abs(1 - y_new) < delta:
                    y_new = 0.0
                if abs(1 - z_new) < delta:
                    z_new = 0.0
                found = False
                for x0, y0, z0 in zip(xs, ys, zs):
                    if (
                        abs(x_new - x0) < delta
                        and abs(y_new - y0) < delta
                        and abs(z_new - z0) < delta
                    ):
                        found = True
                        break
                if not found:
                    xs.append(x_new)
                    ys.append(y_new)
                    zs.append(z_new)
                    symbols.append(symbol)
        self.atoms.append(x=xs, y=ys, z=zs, symbol=symbols)

    def to_mmcif_text(self):
        """Create the text of a mmCIF file from this configuration.

        Returns
        -------
        text : str
            The text of the file.
        """

        lines = []

        atoms = self.atoms
        bonds = self.bonds

        # Get the chemical formula
        formula, empirical_formula, Z = self.formula

        _id = empirical_formula.replace(" ", "")

        # And created the file, line-by-line
        lines = []
        lines.append("# Generated by MolSSI SEAMM")
        lines.append(f"data_{_id}")
        lines.append(f"_chem_comp.name '{formula}'")
        lines.append(f"_chem_comp.id '{_id}'")
        lines.append(f"_chem_comp.formula   '{formula}'")

        # Cell information
        if self.periodicity == 3:
            cell = self.cell
            a, b, c, alpha, beta, gamma = cell.parameters
            volume = cell.volume
            # lines.append("_symmetry_space_group_name_H-M   'P 1'")
            lines.append(f"_cell.length_a   {a}")
            lines.append(f"_cell.length_b   {b}")
            lines.append(f"_cell.length_c   {c}")
            lines.append(f"_cell.angle_alpha   {alpha}")
            lines.append(f"_cell.angle_beta    {beta}")
            lines.append(f"_cell.angle_gamma   {gamma}")
            # lines.append("_symmetry_Int_Tables_number   1")
            lines.append(f"_cell.volume   {volume}")
            lines.append(f"_cell.formula_units_Z   {Z}")
            # lines.append("loop_")
            # lines.append(" _symmetry_equiv_pos_site_id")
            # lines.append(" _symmetry_equiv_pos_as_xyz")
            # lines.append("  1  'x, y, z'")

        # lines.append(f"_chemical_formula_structural   '{empirical_formula}'")
        lines.append(f"_chemical_formula.sum   '{formula}'")

        # The atoms
        lines.append("loop_")
        lines.append(" _chem_comp_atom.comp_id")
        lines.append(" _chem_comp_atom.atom_id")
        lines.append(" _chem_comp_atom.type_symbol")
        lines.append(" _chem_comp_atom.model_Cartn_x")
        lines.append(" _chem_comp_atom.model_Cartn_y")
        lines.append(" _chem_comp_atom.model_Cartn_z")
        lines.append(" _chem_comp_atom.pdbx_model_Cartn_x_ideal")
        lines.append(" _chem_comp_atom.pdbx_model_Cartn_y_ideal")
        lines.append(" _chem_comp_atom.pdbx_model_Cartn_z_ideal")
        lines.append(" _chem_comp_atom.pdbx_component_comp_id")
        lines.append(" _chem_comp_atom.pdbx_residue_numbering")

        # Need unique names
        if "names" in atoms:
            original_names = atoms.get_column("names")
        else:
            original_names = atoms.symbols

        names = []
        tmp = {}
        for name in original_names:
            if name in tmp:
                tmp[name] += 1
                names.append(name + str(tmp[name]))
            else:
                tmp[name] = 1
                names.append(name)

        XYZ = atoms.get_coordinates(in_cell="molecule", fractionals=False)
        XYZa = atoms.get_coordinates(fractionals=False)
        XYZa = XYZ

        symbols = atoms.symbols
        for element, name, xyza, xyz in zip(symbols, names, XYZa, XYZ):
            xa, ya, za = xyza
            x, y, z = xyz
            lines.append(
                f"MOL1 {name} {element} {xa:.3f} {ya:.3f} {za:.3f} "
                f"{x:.3f} {y:.3f} {z:.3f} HET 1"
            )

        # The bonds
        lines.append("#")
        lines.append("loop_")
        lines.append(" _chem_comp_bond.comp_id")
        lines.append(" _chem_comp_bond.atom_id_1")
        lines.append(" _chem_comp_bond.atom_id_2")
        lines.append(" _chem_comp_bond.value_order")
        index = {j: i for i, j in enumerate(atoms.ids)}
        for row in bonds.bonds():
            i = index[row["i"]]
            j = index[row["j"]]
            order = bond_order[row["bondorder"]]
            nm1 = names[i]
            nm2 = names[j]
            lines.append(f"MOL1 {nm1} {nm2} {order}")

        # And that is it!
        return "\n".join(lines)

    def from_mmcif_text(self, text):
        """Create this configuration from an MMCIF file..

        Parameters
        ----------
        text : str
            The text from the MMCIF file

        Returns
        -------
        None

        Notes
        -----
        This can be called from a `SystemDB`, `_System` or `_Configuration`
        object. The behavior and errors differ depending on what type of object
        is calling it:

            `SystemDB`

            When called from a `_System` object, a new `_System` will be
            created for each datablock in the CIF data, and a configuration
            will be created to hold the structure, unless their is an NMR
            ensemble, in which case each structure in the ensemble will be
            placed in a different configuration.

            `_System`

            In this case it is an error if there is more than one datablock.
            A new configuration will be created with the structure in the CIF
            datablock, unless the CIF data contains an NMR ensemble, in which
            case a configuration will be added for each conformer.

            `_Configuration`

            It is an error if there is more than one datablock in the CIF data.
            The configuration will be cleared and the structure from CIF data
            inserted into it. If there is an NMR ensemble in the datablock, and
            the representative conformer is identified, it will be loaded into
            the configuration. Otherwise an error will be raised.
        """
        # What type of object is this?
        if "SystemDB" in str(type(self)):
            my_class = "SystemDB"
        elif "_System" in str(type(self)):
            my_class = "System"
        elif "_Configuration" in str(type(self)):
            my_class = "Configuration"

        cif = CifFile.ReadCif(io.StringIO(text))

        if my_class != "SystemDB" and len(cif) != 1:
            raise RuntimeError(f"There are {len(cif)} data blocks in the mmcif file.")

        # Loop over the datablocks, processing as we go.
        systems = []
        configurations = {}
        for name, data in cif.items():
            # Check the sanity of the input
            coordinate_system = "Cartesian"
            if "_atom_site.id" in data:
                if "_chem_comp_atom.atom_id" in data:
                    raise ValueError("Have both atom_site and chem_comp atoms")

                # Fractional or Cartesian coordinates?
                if "_atom_site.fract_x" in data:
                    coordinate_system = "fractional"
                elif "_atom_site.Cartn_x" not in data:
                    raise KeyError("Couldn't find coordinates in atom_site data")
                elif "_chem_comp_atom.atom_id" in data:
                    if "_chem_comp_atom.model_Cartn_x" not in data:
                        raise KeyError(
                            "Couldn't find coordinates in chem_comp_atom data"
                        )

            # Get the name of the system/configuration
            for key in [
                "_entry.id",
                "_chem_comp.id",
                "_chem_comp.name",
                "_chem_comp.three_letter_code",
            ]:
                if key in data:
                    name = data[key]
                    break

            key = "_pdbx_nmr_ensemble.conformers_submitted_total_number"
            ensemble = key in data
            if ensemble:
                n_ensemble = int(data[key])
                key = "_pdbx_nmr_representative.conformer_id"
                configuration_name = "model_1"
                if key in data:
                    representative = int(data[key])
                    if representative == 1:
                        configuration_name == "representative"
                else:
                    representative = None
            else:
                configuration_name = name
            if my_class == "SystemDB":
                system = self.create_system(name)
                systems.append(system)
                configuration = system.create_configuration(configuration_name)
                configurations[system.id] = [configuration]
            elif my_class == "System":
                system = self
                configuration = system.create_configuration(configuration_name)
                configurations[system.id] = [configuration]
            else:
                configuration = self
                # Reset the configuration
                configuration.clear()
                configuration.periodicity = 0
                configuration.coordinate_system = "Cartesian"
                configuration.name = configuration_name

            # Add the atoms
            atoms = configuration.atoms
            kwargs = {}

            # Get the data from the CIF file, creating columns if needed.
            for cif_key, key, _type, default in [
                ("_chem_comp_atom.atom_id", "name", "str", ""),
                ("_chem_comp_atom.alt_atom_id", None, "str", ""),
                ("_chem_comp_atom.type_symbol", "symbol", "str", ""),
                ("_chem_comp_atom.model_Cartn_x", "x", "float", 0.0),
                ("_chem_comp_atom.model_Cartn_y", "y", "float", 0.0),
                ("_chem_comp_atom.model_Cartn_z", "z", "float", 0.0),
                ("_chem_comp_atom.charge", "formal_charge", "int", 0),
                ("_chem_comp_atom.pdbx_align", None, "int", 0),
                ("_chem_comp_atom.pdbx_aromatic_flag", None, "bool", False),
                ("_chem_comp_atom.pdbx_leaving_atom_flag", None, "bool", False),
                ("_chem_comp_atom.pdbx_stereo_config", None, "bool", False),
                ("_chem_comp_atom.pdbx_component_atom_id", None, "bool", False),
                ("_atom_site.group_PDB", None, "str", "HETATOM"),
                ("_atom_site.id", None, "str", None),
                ("_atom_site.type_symbol", "symbol", "str", None),
                ("_atom_site.label_atom_id", None, "str", None),
                ("_atom_site.label_alt_id", None, "str", ""),
                ("_atom_site.label_comp_id", None, "str", None),
                ("_atom_site.label_asym_id", None, "str", None),
                ("_atom_site.label_entity_id", None, "str", None),
                ("_atom_site.label_seq_id", None, "int", None),
                ("_atom_site.pdbx_PDB_ins_code", None, "str", None),
                ("_atom_site.pdbx_PDB_model_num", "ignore", "int", None),
                ("_atom_site.Cartn_x", "x", "float", 0.0),
                ("_atom_site.Cartn_y", "y", "float", 0.0),
                ("_atom_site.Cartn_z", "z", "float", 0.0),
                ("_atom_site.fract_x", "x", "float", 0.0),
                ("_atom_site.fract_y", "y", "float", 0.0),
                ("_atom_site.fract_z", "z", "float", 0.0),
                ("_atom_site.occupancy", "occupancy", "float", 0.0),
                ("_atom_site.B_iso_or_equiv", None, "float", 0.0),
                ("_atom_site.pdbx_formal_charge", "formal_charge", "int", 0),
            ]:
                if cif_key in data:
                    if key is None:
                        key = cif_key
                    if key == "ignore":
                        key = cif_key
                    elif key not in atoms and key != "symbol":
                        atoms.add_attribute(key, _type, default=default)

                    if _type == "bool":
                        kwargs[key] = [
                            None if x in ".?" else x == "N" for x in data[cif_key]
                        ]
                    elif _type == "int":
                        kwargs[key] = [
                            None if x in ".?" else int(x) for x in data[cif_key]
                        ]
                    elif _type == "float":
                        kwargs[key] = [
                            None if x in ".?" else float(x) for x in data[cif_key]
                        ]
                    else:
                        kwargs[key] = [None if x in ".?" else x for x in data[cif_key]]

            if ensemble:
                if "_atom_site.pdbx_PDB_model_num" not in kwargs:
                    raise KeyError("Is an ensemble but no model numbers.")
                model_num = kwargs.pop("_atom_site.pdbx_PDB_model_num")
                first = 0
                last = first
                model = model_num[0]
                for i in model_num:
                    if i == model:
                        last += 1
                    else:
                        atoms = configuration.atoms
                        model_args = {k: v[first:last] for k, v in kwargs.items()}
                        atom_ids = atoms.append(**model_args)

                        # Make the next configuration and reset pointers
                        model = i
                        first = last
                        last += 1
                        if model == representative:
                            configuration_name = "representative"
                        else:
                            configuration_name = f"model_{model}"
                        configuration = system.create_configuration(configuration_name)
                        configurations[system.id].append(configuration)
                # Put in the last model
                atoms = configuration.atoms
                model_args = {k: v[first:last] for k, v in kwargs.items()}
                atom_ids = atoms.append(**model_args)
                if model != n_ensemble:
                    self.logger.warning(
                        f"The actual number of models ({model}) does not "
                        f"match the claimed number ({n_ensemble})."
                    )
            else:
                if coordinate_system != "Cartesian":
                    raise NotImplementedError("Can't handle fractional coords")

                atom_ids = atoms.append(**kwargs)

                if "_chem_comp_bond.atom_id_1" in data:
                    # Prepare for the bonds, which are labeled by atom name
                    atom_id = {name: i for i, name in zip(atom_ids, kwargs["name"])}

                    # And the bonds
                    bonds = configuration.bonds
                    kwargs = {}
                    for cif_key, key, _type, default in [
                        ("_chem_comp_bond.atom_id_1", "i", "str", ""),
                        ("_chem_comp_bond.atom_id_2", "j", "str", ""),
                        ("_chem_comp_bond.value_order", "bondorder", "int", 1),
                        ("_chem_comp_bond.comp_id", None, "str", None),
                        ("_chem_comp_bond.aromatic_flag", None, "bool", False),
                        ("_chem_comp_bond.stereo_flag", None, "str", None),
                        ("_chem_comp_bond.value_dist", None, "float", None),
                        ("_chem_comp_bond.value_dist_esd", None, "float", None),
                    ]:
                        if cif_key in data:
                            if key is None:
                                key = cif_key
                            if key == "ignore":
                                key = cif_key
                            elif key not in bonds:
                                bonds.add_attribute(key, _type, default=default)

                            if key == "i" or key == "j":
                                kwargs[key] = [atom_id[x] for x in data[cif_key]]
                            elif key == "bondorder":
                                kwargs[key] = [to_bond_order[x] for x in data[cif_key]]
                            elif _type == "bool":
                                kwargs[key] = [
                                    None if x in ".?" else x == "N"
                                    for x in data[cif_key]
                                ]
                            elif _type == "int":
                                kwargs[key] = [
                                    None if x in ".?" else int(x) for x in data[cif_key]
                                ]
                            elif _type == "float":
                                kwargs[key] = [
                                    None if x in ".?" else float(x)
                                    for x in data[cif_key]
                                ]
                            else:
                                kwargs[key] = [
                                    None if x in ".?" else x for x in data[cif_key]
                                ]

                    bonds.append(**kwargs)
        return systems, configurations



# -*- coding: utf-8 -*-

"""A dictionary-like object for holding a system
"""

from collections.abc import MutableMapping
import logging
import pprint  # noqa: F401

from .cif import CIFMixin
from .configuration import _Configuration
from .table import _Table

logger = logging.getLogger(__name__)


class _System(CIFMixin, MutableMapping):
    """A single system -- molecule, crystal, etc. -- in SEAMM.

    Based on a SQLite database, this class provides a general
    datastructure for storing a 'system', which is an atom, molecule,
    crystal or a collection of any number of such components all held in
    a single system, in analogy to a thermodynamic system, which is what
    an experiment is perfromed on.

    The system can hold multiple different configurations of the atoms,
    such as found in an MD trajectory or conformer search. The number of
    atoms may change across configurations as may the bonding, so the
    system is capabale of handling a trajectory for e.g. a grand
    canonical simulation or simulating desposition on a surface. The
    system also can contain bonds as are used in classical valence
    forcefields, and can also cope with changing bonds in e.g. reactive
    forcefields.

    The system class also supports the concept of template molecules or
    subsets, which can be used to define molecules, residues, etc. which
    are then repeated in the full structure. Templates make the process
    of building complex, repeated structures simpler and more efficient,
    and also help track the individual components. There is a lighter
    variation of templates called subsets which do not define a
    substructure, but can label and track such substructures. Atoms may
    belong to more than one subset, or to none other than the special
    subset 'all' which contains all the atoms in a given configuration.

    The key tables are:

    system -- a container for all the components, plus storage for the
        system-wide parameters and options.

    configuration -- a list of configurations in e.g. a trajectory.
    configuration_subset -- connnects configurations with subsets.

    cell -- the information about the periodicity of a configuration, if
        needed.

    symmetry -- the point or space group symmetry of the configuration.
    symop -- the symmetry operations for the group.

    template -- a list of all templates.

    subset -- an instantiation of a template, connected with one or more
        configurations of the system.
    subset_atom -- connects the subset to the atoms, and optionally connects
        the atom to a template atom.

    atom -- the nonvarying part of the description of the atoms
    coordinates -- the varying part of the description of atoms

    element -- basic information about the elements. The atomic number
        (atno) is used for foreign keys in other tables.

    This system class provides an anchor for the other component and
    handles system-wide information such as the periodicity and the
    coordinate system used.

    It also tracks the current configuration and template, which are
    conveniences to avoid having to specify the configuration or
    template for every access to the data.

    There is one special, required subset 'all' that contains all of the
    atoms in each configuration of the system. If there are differing
    numbers or types of atoms in the system, or if the bonding changes,
    there will be multiple versions of the subset, each corresponding to
    a different set of atoms or bonds.

    One or more configurations are connected with each 'all' subset,
    which is how the atoms and bonding are connected to the
    configurations.

    :meta public:
    """

    def __init__(self, system_db, _id, logger=logger):
        self._system_db = system_db
        self._id = _id
        self.logger = logger
        self._current_configuration_id = None  # The current configuration
        self._checkpoints = []
        self._items = {}

    def __del__(self):
        """Delete all instance variables that are objects."""
        del self._system_db

    def __getitem__(self, key):
        """Allow [] access to the dictionary!

        Because some of the items, such as template contain state,
        we need to ensure that the same object is used everywhere. Hence
        the self._items array to store the instances.

        Parameters
        ----------
        key : str
            The table name or name of a multi-table item like atoms.

        Returns
        -------
        table : Table or Atom, Template, Templateatoms, ...
        """
        if key not in self._items:
            self._items[key] = _Table(self, key)
        return self._items[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        raise NotImplementedError(f"Table '{key}' cannot be created yet")

    def __delitem__(self, key):
        """Allow deletion of keys"""
        if key in self:
            self.cursor.execute(f"DROP TABLE '{key}'")

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self.list())

    def __len__(self):
        """The len() command"""
        self.cursor.execute(
            "SELECT COUNT(*)" "  FROM sqlite_master" " WHERE type = 'table'"
        )
        return self.cursor.fetchone()[0]

    # def __repr__(self):
    #     """The string representation of this object"""
    #     raise NotImplementedError()

    # def __str__(self):
    #     """The pretty string representation of this object"""
    #     raise NotImplementedError()

    def __contains__(self, item):
        """Return a boolean indicating if a key exists."""
        # Normal the tablename is used as an identifier, so is quoted with ".
        # Here we need it as a string literal so strip any quotes from it.

        tmp_item = item.strip('"')
        self.cursor.execute(
            "SELECT COUNT(*)"
            "  FROM sqlite_master"
            f" WHERE type = 'table' and name = '{tmp_item}'"
        )
        return self.cursor.fetchone()[0] == 1

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        if self is other:
            return True

        tables = set(self.list())
        other_tables = set(other.list())

        added = tables - other_tables
        deleted = other_tables - tables
        in_common = tables & other_tables

        if len(added) > 0:
            return False
        if len(deleted) > 0:
            return False

        # Need the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.
        name = self.name
        other_name = other.name
        detach = False
        if name != other_name and not self.is_attached(other_name):
            # Attach the other system in order to do comparisons.
            self.attach(other)
            detach = True

        # Check the tables in both systems
        result = True
        for table in in_common:
            if _Table(self, table) != _Table(other, table):
                result = False
                break

        # Detach the other database if needed
        if detach:
            self.detach(other)

        return result

    @property
    def configuration(self):
        """The configuration object for the current configuration."""
        if self.n_configurations == 0:
            return None
        if self._current_configuration_id is None:
            self._current_configuration_id = self.configuration_ids[-1]
        return _Configuration(
            _id=self._current_configuration_id, system_db=self.system_db
        )

    @configuration.setter
    def configuration(self, value):
        if isinstance(value, _Configuration):
            value = value.id
        if value not in self.configuration_ids:
            raise KeyError(f"configuration '{value}' does not exist.")
        self._current_configuration_id = value

    @property
    def configuration_names(self):
        """Get the names of the configurations.

        Returns
        -------
        [str]
            The names of the configurations.
        """
        # The name of the configuration ... find it.
        sql = "SELECT name FROM configuration WHERE system = ?"
        return [x[0] for x in self.db.execute(sql, (self.id,))]

    @property
    def configurations(self):
        """The list of configuration objects for this system."""
        result = []
        for _id in self.configuration_ids:
            result.append(_Configuration(system_db=self.system_db, _id=_id))
        return result

    @property
    def configuration_ids(self):
        """The list of configuration ids."""
        result = []
        sql = "SELECT id FROM configuration WHERE system = ?"
        for row in self.db.execute(sql, (self.id,)):
            result.append(row[0])
        return result

    @property
    def cursor(self):
        return self.system_db.cursor

    @property
    def db(self):
        return self.system_db.db

    @property
    def id(self):
        """The id of this system."""
        return self._id

    @property
    def name(self):
        """Return the name of this system."""
        self.cursor.execute("SELECT name FROM system WHERE id = ?", (self.id,))
        result = self.cursor.fetchone()
        if result is None:
            return None
        else:
            return result[0]

    @name.setter
    def name(self, value):
        self.db.execute("UPDATE system SET name = ? WHERE id = ?", (value, self.id))
        self.db.commit()

    @property
    def n_configurations(self):
        """The number of configurations of the system."""
        sql = "SELECT COUNT(*) FROM configuration WHERE system = ?"
        self.cursor.execute(sql, (self.id,))
        return self.cursor.fetchone()[0]

    @property
    def system_db(self):
        """Return the SystemDB object that contains this system."""
        return self._system_db

    def copy_configuration(
        self,
        configuration=None,
        name=None,
    ):
        """Add a new configuration by copying another configuration.

        Parameters
        ----------
        configuration : int = None
            The configuration to copy. Defaults to the current configuration.
        name : str = None
            A textual name for the configuration (optional)

        Returns
        -------
        cid : int
            The id of the new configuration.

        """
        configuration = self.get_configuration(configuration)

        cid = self["configuration"].append(
            system=self.id,
            name=name,
            periodicity=configuration.periodicity,
            coordinate_system=configuration.coordinate_system,
            symmetry=configuration.symmetry_id,
            cell=configuration.cell_id,
            atomset=configuration.atomset,
            bondset=configuration.bondset,
        )[0]

        return cid

    def create_configuration(
        self,
        name="",
        periodicity=0,
        coordinate_system=None,
        symmetry=None,
        cell_id=None,
        atomset=None,
        bondset=None,
        make_current=True,
    ):
        """Add a new configuration to the system.

        Parameters
        ----------
        name : str = None
            A textual name for the configuration (optional)
        periodicity : int = 0
            The periodicity, 0, or 3 for 0-D or 3-D at the moment
        coordinate_system : str = None
            The coordinate system, 'Cartesian' or 'fractional', to use.
            Defaults to Cartesian for molecules and fractional for crystals.
        symmetry : int or str = None
            The id or name of the point or space group (optional)
        cell_id : int = None
            The id of the _Cell object
        atomset : int = None
            The set of atoms in this configurationn
        bondset : int = None
            The bonds in this configuration
        make_current : bool = True
            If True, make the current configuration.

        Returns
        -------
        _Configuration
            The new configuration.
        """
        kwargs = {}
        if name is not None:
            kwargs["name"] = name
        if periodicity != 0:
            if coordinate_system is None:
                kwargs["coordinate_system"] = "fractional"
            else:
                if coordinate_system.lower()[0] == "c":
                    kwargs["coordinate_system"] = "Cartesian"
                else:
                    kwargs["coordinate_system"] = "fractional"
        if symmetry is not None:
            kwargs["symmetry"] = symmetry
        if cell_id is not None:
            kwargs["cell"] = cell_id
        if atomset is not None:
            kwargs["atomset"] = atomset
        if bondset is not None:
            kwargs["bondset"] = bondset

        cid = self["configuration"].append(
            system=self.id, periodicity=periodicity, **kwargs
        )[0]
        configuration = _Configuration(_id=cid, system_db=self.system_db)

        # If the atomset was given, need to create dummy coordinates
        if atomset is not None:
            n_atoms = configuration.n_atoms
            data = {"configuration": configuration.id}
            data["atom"] = configuration.atoms.ids
            data["x"] = [0.0] * n_atoms
            data["y"] = [0.0] * n_atoms
            data["z"] = [0.0] * n_atoms
            table = _Table(self.system_db, "coordinates")
            table.append(n=n_atoms, **data)

        if make_current:
            self._current_configuration_id = cid

        return configuration

    def create_table(self, name, cls=_Table, other=None):
        """Create a new table with the given name.

        Parameters
        ----------
        name : str
            The name of the new table.

        cls : _Table subclass
            The class of the new table, defaults to _Table

        Returns
        -------
        table : class _Table
            The new table
        """
        if name in self:
            raise KeyError(f"'{name}' already exists in the system.")

        self._items[name] = cls(self, name, other)
        return self._items[name]

    def diff(self, other):
        """Differences between this system and another."""
        result = {}

        if self is other:
            return result

        tables = set(self.list())
        other_tables = set(other.list())

        added = tables - other_tables
        deleted = other_tables - tables
        in_common = tables & other_tables

        if len(added) > 0:
            result["tables added"] = added
        if len(deleted) > 0:
            result["tables deleted"] = deleted

        # Need the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.
        name = self.nickname
        other_name = other.nickname
        detach = False
        if name != other_name and not self.is_attached(other_name):
            # Attach the other system in order to do comparisons.
            self.attach(other)
            detach = True

        # Check the tables in both systems
        for table in in_common:
            table1 = _Table(self, table)
            table2 = _Table(other, table)
            tmp = table1.diff(table2)
            if len(tmp) > 0:
                result[f"table '{table}' diffs"] = tmp

        # Detach the other database if needed
        if detach:
            self.detach(other)

        return result

    def get_configuration(self, cid):
        """Get the specified configuration object.

        Parameters
        ----------
        cid : int or str
            The id (int) or name (str) of the configuration

        Returns
        -------
        _Configuration
            The requested configuration.

        Raises
        ------
        ValueError
            If the configuration does not exist, or more than one have the
            requested name.
        """
        if isinstance(cid, str):
            # The name of the configuration ... find it.
            cid = self.get_configuration_id(cid)

        return _Configuration(_id=cid, system_db=self.system_db)

    def get_configuration_id(self, name):
        """Get the id of the specified configuration.

        Parameters
        ----------
        name : str
            The name of the configuration

        Returns
        -------
        int
            The id of the requested configuration.

        Raises
        ------
        ValueError
            If the configuration does not exist, or more than one have the
            requested name.
        """
        # The name of the configuration ... find it.
        sql = "SELECT id FROM configuration WHERE system = ? AND name = ?"
        ids = [x[0] for x in self.db.execute(sql, (self.id, name))]
        if len(ids) == 0:
            raise ValueError(f"The configuration '{name}' does not exist.")
        elif len(ids) > 1:
            raise ValueError(f"There is more than one configuration named '{name}'")
        return ids[0]

    def list(self):
        """Return a list of all the tables in the system."""
        result = []
        for row in self.db.execute(
            "SELECT name" "  FROM sqlite_master" " WHERE type = 'table'"
        ):
            result.append(row["name"])
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for properties."""

from pathlib import Path  # noqa: F401
import pprint  # noqa: F401
import pytest  # noqa: F401
import time

from molsystem import SystemDB  # noqa: F401

nsystems = 1000000
nproperties = 10


@pytest.fixture(scope="module")
def db_file(tmp_path_factory):
    """Create and temporary directory, and clean it up at the end."""
    dirpath = tmp_path_factory.mktemp("data")
    path = dirpath / "seamm.db"
    db = SystemDB(filename=str(path))
    db.close()

    print(f"temp file: {path}")

    yield str(path)

    path.unlink()
    dirpath.rmdir()


@pytest.fixture()
def disk_db(db_file):
    """Create and return a SystemDB object with the database file."""
    db = SystemDB(filename=db_file)

    yield db

    db.close()


@pytest.fixture()
def saved_db():
    """Create and return a SystemDB object with the database file."""
    db = SystemDB(filename="seamm.db")

    yield db

    db.close()


@pytest.fixture()
def properties(disk_db):
    """Lots of configurations with many properties."""
    # path = Path("seamm.db")
    # exists = path.exists()

    # db = SystemDB(filename="seamm.db")
    exists = False
    db = disk_db

    float_data = db["float_data"]
    int_data = db["int_data"]
    str_data = db["str_data"]

    properties = db.properties

    if not exists:
        t0 = time.perf_counter()
        # Create systems
        cids = []
        names = [str(i) for i in range(nsystems)]
        systems = db["system"].append(name=names)
        cids = db["configuration"].append(system=systems, name="1")

        t1 = time.perf_counter()
        print(f"  {nsystems} systems took {t1-t0:.3f} s")

        # And properties
        fstart = -nproperties
        delta = 10.0 / nsystems
        istart = -nsystems
        for i in range(nproperties):
            fid = properties.add(f"float_{i}", "float")
            iid = properties.add(f"int_{i}", "int")
            sid = properties.add(f"str_{i}", "str")

            fval = fstart + i
            ival = istart + nsystems * i
            fvals = []
            ivals = []
            svals = []
            for cid in cids:
                fvals.append(fval)
                ivals.append(ival)
                svals.append(f"string {ival}")
                fval += delta
                ival += 1
            float_data.append(configuration=cids, property=fid, value=fvals)
            int_data.append(configuration=cids, property=iid, value=ivals)
            str_data.append(configuration=cids, property=sid, value=svals)

        t1 = time.perf_counter()
        print(f"  db creation took {t1-t0:.3f} s")

    yield properties

    db.close()


@pytest.mark.timing
def test_query2(properties):
    """Test a simple query"""

    t0 = time.perf_counter()
    result = properties.query(
        "float_0",
        "between",
        -10.0,
        -10.0 + 1000 / nsystems,
        what=["float_8", "str_9"],
    )
    t1 = time.perf_counter()
    print(f"  test took {t1-t0:.6f} s")
    print(f"        retrieved {len(result['str_9'])} strings")
    print(f"              and {len(result['float_8'])} floats")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for configurations."""

import pytest  # noqa: F401


def test_formula(AceticAcid):
    """Test the formula generation."""
    formula, empirical_formula, Z = AceticAcid.formula
    assert formula == "C2 H4 O2"
    assert empirical_formula == "C H2 O"
    assert Z == 2


def test_formula_periodic(vanadium):
    """Test the formula generation."""
    formula, empirical_formula, Z = vanadium.formula
    assert "".join(formula) == "V2"
    assert "".join(empirical_formula) == "V"
    assert Z == 2


def test_density(vanadium):
    """Test the density, and implicitly the mass and volume."""

    assert abs(vanadium.density - 6.0817308915133) < 1.0e-06



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the topology mixin of the system class."""

import pprint

import pytest  # noqa: F401


def test_bonded_neighbors(AceticAcid):
    """Test the generation of bonded neighbors."""
    result = {
        1: [2, 3, 4, 5],
        2: [1],
        3: [1],
        4: [1],
        5: [1, 6, 7],
        6: [5],
        7: [5, 8],
        8: [7],
    }

    neighbors = AceticAcid.bonded_neighbors()
    if neighbors != result:
        pprint.pprint(neighbors)
    assert neighbors == result


def test_molecules(AceticAcid):
    """Test the finding molecules in the system."""
    result = [[1, 2, 3, 4, 5, 6, 7, 8]]

    molecules = AceticAcid.find_molecules()
    if molecules != result:
        pprint.pprint(molecules)
    assert molecules == result


def test_multiple_molecules(CH3COOH_3H2O):
    """Test the finding molecules in the system."""
    result = [[1, 2, 3, 4, 5, 6, 7, 8], [9, 10, 11], [12, 13, 14], [15, 16, 17]]

    molecules = CH3COOH_3H2O.find_molecules()
    if molecules != result:
        pprint.pprint(molecules)
    assert molecules == result


def test_molecule_templates(disordered):
    """Test making templates for the molecules."""
    result_tids = [1]

    configuration = disordered
    templates = configuration.create_molecule_templates(create_subsets=False)
    tids = [x.id for x in templates]
    if tids != result_tids:
        print("tids")
        pprint.pprint(tids)
    assert tids == result_tids


def test_molecule_subsets(disordered):
    """Test making templates and subsets for the molecules."""
    result_tids = [1]
    result_sids = {1: [1, 2]}

    configuration = disordered
    templates, subsets = configuration.create_molecule_templates()
    tids = [x.id for x in templates]
    if tids != result_tids:
        print("tids")
        pprint.pprint(tids)
    sids = {t: [x.id for x in s] for t, s in subsets.items()}
    if sids != result_sids:
        print("sids")
        pprint.pprint(sids)
    assert tids == result_tids
    assert sids == result_sids

    # And that the atoms are indeed ordered correctly
    tid = tids[0]
    subset1, subset2 = subsets[tid]
    atnos1 = subset1.atoms.atomic_numbers
    atnos2 = subset2.atoms.atomic_numbers
    if atnos1 != atnos2:
        print(f"atnos1 = {atnos1}")
        print(f"atnos2 = {atnos2}")
    assert atnos1 == atnos2

    coords1 = subset1.atoms.coordinates
    coords2 = subset2.atoms.coordinates

    # At the moment the methyl hydrogens are not properly ordered -- they are
    # equivalent -- so remove them to test.
    c1 = list(coords1)
    del c1[1:4]
    c2 = list(coords2)
    del c2[1:4]

    if c1 != c2:
        print("coords1")
        pprint.pprint(coords1)
        print("coords2")
        pprint.pprint(coords2)
    assert c1 == c2

    assert subset1.atoms.get_n_atoms("atno", "==", 6) == 2

    count = 0
    for row in subset2.atoms.atoms("atno", "==", 6):
        count += 1
    assert count == 2



# -*- coding: utf-8 -*-

"""Functions for handling SMILES"""

import logging

try:
    from openbabel import openbabel
except ModuleNotFoundError:
    print(
        "Please install openbabel using conda:\n"
        "     conda install -c conda-forge openbabel"
    )
    raise

logger = logging.getLogger(__name__)


class SMILESMixin:
    """A mixin for handling SMILES."""

    @property
    def canonical_smiles(self):
        """Return the canonical SMILES string for this object."""
        return self.to_smiles(canonical=True)

    @property
    def smarts(self):
        """Return the SMARTS string for this object."""
        return self.to_smarts()

    @property
    def smiles(self):
        """Return the SMILES string for this object."""
        return self.to_smiles()

    def to_smiles(self, canonical=False, hydrogens=False):
        """Create the SMILES string from the system.

        Parameters
        ----------
        canonical : bool = False
            Whether to create canonical SMILES
        hydrogens : bool = False
            Whether to keep H's in the SMILES string.

        Returns
        -------
        str
            The SMILES string, or (SMILES, name) if the rname is requested
        """
        logger.info("to_smiles")

        obConversion = openbabel.OBConversion()
        if canonical:
            obConversion.SetOutFormat("can")
        else:
            obConversion.SetOutFormat("smi")

        mol = self.to_OBMol()

        if hydrogens:
            obConversion.AddOption("h")
        smiles = obConversion.WriteString(mol)

        logger.info(f"smiles = '{smiles}'")

        return smiles.strip()

    def from_smiles(self, smiles, name=None):
        """Create the system from a SMILES string.

        Parameters
        ----------
        smiles : str
            The SMILES string
        name : str = None
            The name of the molecule

        Returns
        -------
        None
        """

        save = self.name

        obConversion = openbabel.OBConversion()
        obConversion.SetInAndOutFormats("smi", "mdl")
        obConversion.AddOption("3")
        mol = openbabel.OBMol()
        obConversion.ReadString(mol, smiles)

        # Add hydrogens
        mol.AddHydrogens()

        # Get coordinates for a 3-D structure
        builder = openbabel.OBBuilder()
        builder.Build(mol)

        # molfile = obConversion.WriteString(mol)
        # self.from_molfile_text(molfile)

        self.from_OBMol(mol)

        if name is not None:
            self.name = name
        else:
            self.name = save

    def to_smarts(self):
        """Generate a SMARTS string for this object.

        Returns
        -------
        str
            The SMARTS string.
        """

        generator = GenSMARTS(self)
        return generator.smarts


class GenSMARTS(object):
    """A class to generate SMARTS strings for an object.

    Parameters
    ----------
    mol_object : _Configuration, _Template, _Subset
    """

    def __init__(self, mol_object=None):
        self.smarts = ""
        self._mol_object = None
        self.symbols = []
        self.visited = []
        self.neighbors = []

        self.mol_object = mol_object

    @property
    def mol_object(self):
        """The molecular object to work with."""
        return self._mol_object

    @mol_object.setter
    def mol_object(self, value):
        self._mol_object = value

        if self._mol_object is not None:
            atoms = self.mol_object.atoms
            bonds = self.mol_object.bonds

            self.symbols = atoms.symbols

            # Find the neighbors of each atom for walking the structure
            tmp_neighbors = {i: [] for i in atoms.ids}
            for bond in bonds.bonds():
                i = bond["i"]
                j = bond["j"]
                bo = bond["bondorder"]
                tmp_neighbors[i].append((j, bo))
                tmp_neighbors[j].append((i, bo))

            # The index of each atom id in the list, 0..natoms-1
            index = {j: i for i, j in enumerate(atoms.ids)}

            # Convert to indices
            self.neighbors = {}
            for i, js in tmp_neighbors.items():
                self.neighbors[index[i]] = [(index[j], bo) for j, bo in js]

            # An array to keep track of the atoms visited
            self.visited = [False] * atoms.n_atoms

            # And generate the SMARTS string
            self.smarts = self._recurse(atom=0)

    def _recurse(self, atom):
        """Helper method for recursion to generate SMARTS.

        Parameters
        ----------
        atom : int
            Index of the atom to work with.
        """
        symbol = self.symbols[atom]
        self.visited[atom] = True

        # How many neighbors, and how many are hydrogens
        nX = len(self.neighbors[atom])
        nH = 0
        for neighbor, bondorder in self.neighbors[atom]:
            if self.symbols[neighbor] == "H":
                nH += 1

        # Find any non-hydrogen neighbors that have not been visited,
        # and recurse.
        sub_smarts = []
        for neighbor, bondorder in self.neighbors[atom]:
            if self.symbols[neighbor] != "H" and not self.visited[neighbor]:
                if bondorder == 1:
                    bond = ""
                elif bondorder == 2:
                    bond = "="
                elif bondorder == 3:
                    bond = "#"
                elif bondorder == 5:
                    # Aromatic
                    symbol = symbol.lower()
                    bond = ":"
                else:
                    bond = "~"
                sub_smarts.append(f"{bond}{self._recurse(neighbor)}")

        smarts = f"[{symbol}H{nH}X{nX}]"
        if len(sub_smarts) > 0:
            for sub in sub_smarts[0:-1]:
                smarts += f"({sub})"
            smarts += sub_smarts[-1]

        return smarts



# -*- coding: utf-8 -*-

"""Functions for handling MDL molfiles"""

import logging
import time

from molsystem import elements

logger = logging.getLogger(__name__)


class MolFileMixin:
    """A mixin for handling MDL Molfiles."""

    def to_molfile_text(self, title=None, comment="Exported from SEAMM"):
        """Create the text of the Molfile from the system.

        Parameters
        ----------
        title : str = None
            The title for the structure, by default the system name.
        comment : str = 'Exported from SEAMM'
            Comment line

        Returns
        -------
        text : str
            The text of the file.
        """

        lines = []

        atoms = self.atoms
        bonds = self.bonds

        n_atoms = atoms.n_atoms
        n_bonds = bonds.n_bonds

        nsgroups = 0
        n3d = 0
        is_chiral = 0  # may need to think about this later.

        if title is None:
            lines.append(self.name)
        else:
            lines.append(title)
        date_time = time.strftime("%m%d%y%H%M")

        lines.append("PS" + "SEAMM_WF" + date_time + "3D")
        lines.append(comment)
        lines.append("  0  0  0     0  0            999 V3000")

        lines.append("M  V30 BEGIN CTAB")
        lines.append(
            "M  V30 COUNTS {} {} {} {} {}".format(
                n_atoms, n_bonds, nsgroups, n3d, is_chiral
            )
        )
        lines.append("M  V30 BEGIN ATOM")
        count = 0
        if "formal charges" in atoms:
            for row in atoms.atoms():
                count += 1
                symbol = elements.to_symbols([row["atno"]])[0]
                lines.append(
                    f"M  V30 {count} {symbol} {row['x']} {row['y']} {row['z']}"
                    " 0 CHG={row['formal charge']}"
                )
        else:
            for row in atoms.atoms():
                count += 1
                symbol = elements.to_symbols([row["atno"]])[0]
                lines.append(
                    f"M  V30 {count} {symbol} {row['x']} {row['y']} {row['z']}" " 0"
                )
        lines.append("M  V30 END ATOM")
        lines.append("M  V30 BEGIN BOND")
        count = 0
        for row in bonds.bonds():
            count += 1
            lines.append(f"M  V30 {count} {row['bondorder']} " f"{row['i']} {row['j']}")
        lines.append("M  V30 END BOND")
        lines.append("M  V30 END CTAB")
        lines.append("M  END")

        return "\n".join(lines)

    def from_molfile_text(self, data):
        """Create the system from an MDL Molfile, version 3

        Parameters
        ----------
        data : str
            The complete text of the Molfile.
        """

        self.clear()
        self.periodicity = 0

        n_molecules = 0
        lines = enumerate(data.splitlines())

        # title
        lineno, title = next(lines)
        self.name = title.strip()
        # header
        next(lines)
        # comment
        next(lines)
        lineno, line = next(lines)
        if line.split()[6] != "V3000":
            raise RuntimeError(
                f"molfile:to_seamm -- the file is not version 3: '{line}'"
            )
        for lineno, line in lines:
            logger.debug(f"{lineno}: {line}")
            if "M  END" in line:
                break
            elif "M  V30 BEGIN CTAB" in line:
                n_molecules += 1
                if n_molecules > 1:
                    raise NotImplementedError("Multiple molecules?")
            elif "M V30 END CTAB" in line:
                pass
            elif "M  V30 COUNTS" in line:
                natoms, nbonds, nsgroups, n3d, is_chiral = line.split()[3:]

                natoms = int(natoms)
                nbonds = int(nbonds)
                # not used, yet.
                # nsgroups = int(nsgroups)
                # n3d = int(n3d)
                # is_chiral = bool(is_chiral)
            elif "M  V30 BEGIN ATOM" in line:
                logger.debug("In atom table")
                xs = []
                ys = []
                zs = []
                symbols = []
                formal_charges = []
                have_formal_charges = False
                for lineno, line in lines:
                    if "M  V30 END ATOM" in line:
                        logger.debug(f"Saving {len(xs)} atoms to system")
                        if have_formal_charges and "formal_charge" not in self.atoms:
                            logger.debug("   with formal charges")
                            self.atoms.add_attribute(
                                "formal_charge", coltype="int", default=0
                            )
                            atom_ids = self.atoms.append(
                                x=xs,
                                y=ys,
                                z=zs,
                                symbol=symbols,
                                formal_charge=formal_charges,
                            )
                        else:
                            atom_ids = self.atoms.append(
                                x=xs, y=ys, z=zs, symbol=symbols
                            )
                        break
                    i, symbol, x, y, z, q = line.split()[2:8]
                    xs.append(float(x))
                    ys.append(float(y))
                    zs.append(float(z))
                    symbols.append(symbol)
                    if "CHG=" in line:
                        for tmp in line.split()[8:]:
                            if "CHG=" in tmp:
                                formal_charges.append(int(tmp[4:]))
                                have_formal_charges = True
                    else:
                        formal_charges.append(0)
            elif "M  V30 BEGIN BOND" in line:
                logger.debug("In bond table")
                iatoms = []
                jatoms = []
                bondorders = []
                for lineno, line in lines:
                    if "M  V30 END BOND" in line:
                        if len(iatoms) > 0:
                            self.bonds.append(
                                i=iatoms,
                                j=jatoms,
                                bondorder=bondorders,
                            )
                        break
                    bondorder, iatom, jatom = line.split()[3:6]
                    iatoms.append(atom_ids[int(iatom) - 1])
                    jatoms.append(atom_ids[int(jatom) - 1])
                    bondorders.append(int(bondorder))



# -*- coding: utf-8 -*-

import csv
import logging
from pathlib import Path
import pkg_resources
import pprint  # noqa: F401

logger = logging.getLogger(__name__)


class _Properties(object):
    """A class for handling the properties in the star schema."""

    def __init__(self, system_db):
        self._system_db = system_db
        self._standard_properties = None

    @property
    def cursor(self):
        return self.system_db.cursor

    @property
    def db(self):
        return self.system_db.db

    @property
    def standard_properties(self):
        """The standard properties recognized by SEAMM.

        These are officially defined properties that can be used anywhere in SEAMM, as
        long as the type and definition correspond to the standard.

        You can create other properties on the fly, but they must be prefixed by a
        unique name followed by a dot ('.') so that they do not conflict with either the
        standard properties or other properties defined on-the-fly. Typically the unique
        name should  that of the program generating the property, which implicitly
        defines details of the property.

        For example, the standard property "enthalpy of formation" refers to the
        experimental heat of formation, or a calculated value comparable to experimental
        values. If you are not sure what the heat of formation in e.g. MOPAC is, you
        could create a new property "MOPAC.enthalpy of formation", which is clearly
        similar to the standard "enthalpy of formation". If the community decides that
        it is indeed the same, it can be replaced by the standard form, and also aliased
        to it for backwards compatibility.
        """
        if self._standard_properties is None:
            self._standard_properties = {}
            path = Path(pkg_resources.resource_filename(__name__, "data/"))
            csv_file = path / "standard_properties.csv"
            with open(csv_file, newline="", encoding="utf-8-sig") as fd:
                data = csv.reader(fd)
                line = 0
                for row in data:
                    line += 1
                    if line == 1:
                        # Headers
                        headers = [*row]
                        if headers != [
                            "Property",
                            "Type",
                            "Units",
                            "Description",
                            "URL",
                        ]:
                            raise ValueError(
                                "Header of standard properties file not valid: "
                                + ", ".join(headers)
                            )
                    else:
                        property = row[0]
                        data = self._standard_properties[property] = {}
                        for key, value in zip(headers[1:], row[1:]):
                            data[key] = value
        return self._standard_properties

    @property
    def system_db(self):
        """Return the SystemDB object."""
        return self._system_db

    def add(self, name, _type="float", units=None, description="", noerror=False):
        """Add a property to the database. By default, it is an error if it already
        exists!

        Parameters
        ----------
        name : str
            The name of the property.
        _type : str = "float"
            The type of the property -- float, int, str. Default is "float".
        units : str = None
            The units of the property value, as a string that Pint can interpret.
        description : str
            A longer, text description of the property.
        noerror: bool = False
            Whether to quietly ignore an existing entry

        Returns
        -------
        int
            The database id of the property.

        Note
        ----
        If the property is registered in the list of standard properties, just
        give the name and the rest will be filled in from the standard property
        metadata.
        """
        if self.exists(name):
            if noerror:
                return self.property_id(name)
            else:
                raise ValueError(f"Property '{name}' already exists.")

        table = self.system_db["property"]
        if name in self.standard_properties:
            data = self.standard_properties[name]
            result = table.append(
                name=name,
                type=data["Type"],
                units=data["Units"],
                description=data["Description"],
            )
        else:
            result = table.append(
                name=name,
                type=_type,
                units=units,
                description=description,
            )
        return result[0]

    def create_schema(self):
        """Add the needed tables to the database."""
        # The property dimension table
        table = self.system_db["property"]
        table.add_attribute("id", coltype="int", pk=True)
        table.add_attribute("name", coltype="str", index="unique")
        table.add_attribute("type", coltype="str")
        table.add_attribute("units", coltype="str")
        table.add_attribute("description", coltype="str")

        # Floating point facts
        table = self.system_db["float_data"]
        table.add_attribute("id", coltype="int", pk=True)
        table.add_attribute("configuration", coltype="int", references="configuration")
        table.add_attribute("property", coltype="int", references="property")
        table.add_attribute("value", coltype="float")

        self.db.execute(
            "CREATE INDEX float_data_idx_configuration_property_value"
            "    ON float_data(configuration, property, value)"
        )
        self.db.execute(
            "CREATE INDEX float_data_idx_property_value"
            "    ON float_data(property, value)"
        )
        self.db.execute(
            "CREATE INDEX float_data_idx_configuration_property"
            "    ON float_data(configuration, property)"
        )

        # Integer facts
        table = self.system_db["int_data"]
        table.add_attribute("id", coltype="int", pk=True)
        table.add_attribute("configuration", coltype="int", references="configuration")
        table.add_attribute("property", coltype="int", references="property")
        table.add_attribute("value", coltype="int")

        self.db.execute(
            "CREATE INDEX int_data_idx_configuration_property_value"
            "    ON int_data(configuration, property, value)"
        )
        self.db.execute(
            "CREATE INDEX int_data_idx_property_value"
            "    ON int_data(property, value)"
        )
        self.db.execute(
            "CREATE INDEX int_data_idx_configuration_property"
            "    ON int_data(configuration, property)"
        )

        # String facts
        table = self.system_db["str_data"]
        table.add_attribute("id", coltype="int", pk=True)
        table.add_attribute("configuration", coltype="int", references="configuration")
        table.add_attribute("property", coltype="int", references="property")
        table.add_attribute("value", coltype="str")

        self.db.execute(
            "CREATE INDEX str_data_idx_configuration_property_value"
            "    ON str_data(configuration, property, value)"
        )
        self.db.execute(
            "CREATE INDEX str_data_idx_property_value"
            "    ON str_data(property, value)"
        )
        self.db.execute(
            "CREATE INDEX str_data_idx_configuration_property"
            "    ON str_data(configuration, property)"
        )

    def exists(self, name):
        """Whether the named property exists.

        Parameters
        ----------
        name : str
            The name of the property.

        Returns
        -------
        bool
            True if the property is registered, False otherwise.
        """
        self.cursor.execute("SELECT COUNT(*) FROM property WHERE name = ?", (name,))
        return self.cursor.fetchone()[0] != 0

    def get(self, configuration_id, _property):
        """Get the given property value for the configuration.

        Parameters
        ----------
        configuration_id : int
            The id of the configuration.
        _property : int or str
            The id or name of the property.

        Returns
        -------
        int, float, or str
            The value of the property.
        """
        if isinstance(_property, str):
            if self.exists(_property):
                pid = self.property_id(_property)
            else:
                raise ValueError(f"Property '{_property}' does not exist.")
        else:
            pid = _property

        ptype = self.property_type(pid)
        sql = (
            f"SELECT value FROM {ptype}_data"
            "  WHERE configuration = ? AND property = ?"
        )
        self.cursor.execute(sql, (configuration_id, pid))
        result = self.cursor.fetchone()
        if result is None:
            raise ValueError(
                f"Property {_property} does not exist for configuration "
                f"{configuration_id}"
            )
        return result[0]

    def known_properties(self):
        """List the known properties."""
        self.cursor.execute("SELECT name FROM property")
        return [row[0] for row in self.cursor.fetchall()]

    def property_id(self, name):
        """The id for a property

        Parameters
        ----------
        name : str
            The name of the property.

        Returns
        -------
        int
            The database id for the property.
        """
        self.cursor.execute("SELECT id FROM property WHERE name = ?", (name,))
        return self.cursor.fetchone()[0]

    def property_name(self, pid):
        """The name of a property

        Parameters
        ----------
        pid : int
            The id of the property.

        Returns
        -------
        str
            The name of the property.
        """
        self.cursor.execute("SELECT name FROM property WHERE id = ?", (pid,))
        return self.cursor.fetchone()[0]

    def property_type(self, _property):
        """The type of a property

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        str
            The type of the property.
        """
        if isinstance(_property, str):
            self.cursor.execute(
                "SELECT type FROM property WHERE name = ?", (_property,)
            )
        else:
            self.cursor.execute("SELECT type FROM property WHERE id = ?", (_property,))
        return self.cursor.fetchone()[0]

    def put(self, configuration_id, _property, value):
        """Store the given property value for the configuration.

        Parameters
        ----------
        configuration_id : int
            The id of the configuration.
        _property : int or str
            The id or name of the property.
        value : int, float, or str
            The value to store.
        """
        if isinstance(_property, str):
            if not self.exists(_property):
                if _property in self.standard_properties:
                    self.add(_property)
                else:
                    raise ValueError(f"Property '{_property}' does not exist.")
            pid = self.property_id(_property)
        else:
            pid = _property

        ptype = self.property_type(pid)
        sql = (
            f"INSERT INTO {ptype}_data (configuration, property, value)"
            "      VALUES(?, ?, ?)"
        )
        self.db.execute(sql, (configuration_id, pid, value))

    def query(self, *args, what=["configuration"]):
        """Find configurations that match the query defined by the args.

        what : [str or int]]
            What to return ... "configuration", property names or ids.
        where : {}

        Note
        ----
        args should be a multiple of 3 in length, with each triplet being
        (property, operator, value). The property
        """
        sql = "SELECT"
        where = "WHERE"
        criteria = []
        tables = {}
        results = []
        n_tables = 0
        for i, item in enumerate(what):
            if i > 0:
                sql += ","
            if isinstance(item, str):
                if item == "configuration":
                    sql += " t0.configuration"
                    results.append("configuration_id")
                else:
                    pid = self.property_id(item)
                    ptype = self.property_type(pid)
                    table = ptype + "_data"
                    if table not in tables:
                        tables[table] = {pid: f"t{n_tables}"}
                        n_tables += 1
                    elif pid not in tables[table]:
                        tables[table][pid] = f"t{n_tables}"
                        n_tables += 1
                    alias = tables[table][pid]
                    sql += f" {alias}.value"
                    criteria.append(f"{alias}.property == {pid}")
                    results.append(item)
            elif isinstance(item, int):
                pid = item
                ptype = self.property_type(pid)
                table = ptype + "_data"
                if table not in tables:
                    tables[table] = {pid: f"t{n_tables}"}
                    n_tables += 1
                elif pid not in tables[table]:
                    tables[table][pid] = f"t{n_tables}"
                    n_tables += 1
                alias = tables[table][pid]
                sql += f" {alias}.value"
                criteria.append(f"{alias}.property == {pid}")
                results.append(self.property_name(pid))

        # And the WHERE clause ...
        where = {0: ""}
        items = iter(args)
        level = 0
        for item in items:
            if item == "(":
                level += 1
                where[level] = " ("
            elif item == ")":
                where[level] += ")"
                where[level - 1] += where[level]
                del where[level]
                level -= 1
            else:
                # Configuration or Property name or id
                if isinstance(item, str):
                    pid = self.property_id(item)
                    ptype = self.property_type(pid)
                    table = ptype + "_data"
                    if table not in tables:
                        tables[table] = {pid: f"t{n_tables}"}
                        n_tables += 1
                    elif pid not in tables[table]:
                        tables[table][pid] = f"t{n_tables}"
                        n_tables += 1
                    alias = tables[table][pid]
                    if where[level] != "" and where[level] != "( ":
                        where[level] += " AND"
                    where[level] += f" {alias}.property == {pid}"
                elif isinstance(item, int):
                    pid = item
                    ptype = self.property_type(pid)
                    table = ptype + "_data"
                    if table not in tables:
                        tables[table] = {pid: f"t{n_tables}"}
                        n_tables += 1
                    elif pid not in tables[table]:
                        tables[table][pid] = f"t{n_tables}"
                        n_tables += 1
                    alias = tables[table][pid]
                    if where[level] != "" and where[level] != "( ":
                        where[level] += " AND"
                    where[level] += f" {alias}.property == {pid}"

                operator = next(items)
                if operator.lower() == "between":
                    if ptype == "str":
                        where[level] += (
                            f' AND {alias}.value BETWEEN "{next(items)}"'
                            f' AND "{next(items)}"'
                        )
                    else:
                        where[level] += (
                            f" AND {alias}.value BETWEEN {next(items)}"
                            f" AND {next(items)}"
                        )
                else:
                    if ptype == "str":
                        where[level] += f' AND {alias}.value {operator} "{next(items)}"'
                    else:
                        where[level] += f" AND {alias}.value {operator} {next(items)}"

        # Put it all together ... FROM clause plus needed criteria
        # if len(tables) == 0:
        #     # No tables ... so probably getting all configurations with properties
        #     tables = ["int_data", "float_data", "str_data"]
        sql += " FROM "

        tmp = []
        for table, data in tables.items():
            for pid, alias in data.items():
                tmp.append(f"{table} AS {alias}")
        sql += ", ".join(tmp)

        # Now for the WHERE clause
        sql += " WHERE "
        tmp = []
        for table, data in tables.items():
            for pid, alias in data.items():
                if alias != "t0":
                    tmp.append(f"{alias}.configuration = t0.configuration")
        sql += " AND ".join(tmp)

        if len(criteria) > 0:
            if len(tmp) > 0:
                sql += " AND "
            sql += " AND ".join(criteria)

        # The user criteria, if any
        if where[0] != "":
            if len(tmp) > 0:
                sql += " AND"
            sql += where[0]

        logger.info(sql)

        result = {item: [] for item in results}
        try:
            self.cursor.execute(sql)
        except Exception:
            print("")
            print(f"{tmp=}")
            print(f"{criteria=}")
            print(f"{where[0]=}")
            raise

        for row in self.cursor:
            for item, value in zip(results, row):
                result[item].append(value)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy
import pytest  # noqa: F401
import os
import os.path
import time

from molsystem import SystemDB  # noqa: F401

"""Performance tests for molsystem."""

natoms = 1000000


@pytest.fixture(scope="module")
def db_file(tmp_path_factory):
    """Create and temporary directory, and clean it up at the end."""
    dirpath = tmp_path_factory.mktemp("data")
    path = dirpath / "seamm.db"
    db = SystemDB(filename=str(path))
    db.close()

    print(f"temp file: {path}")

    yield str(path)

    path.unlink()
    dirpath.rmdir()


@pytest.fixture()
def disk_db(db_file):
    """Create and return a SystemDB object with the database file."""
    db = SystemDB(filename=db_file)

    yield db

    db.close()


@pytest.fixture(scope="module")
def mem_db():
    """Create and return a SystemDB using an in-memory DB"""
    db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")

    yield db

    db.close()


@pytest.mark.timing
def test_creation(disk_db):
    """Create a system with a million atoms"""
    print(f"\nCreating a system with {natoms} atoms")
    t0 = time.perf_counter()
    system = disk_db.create_system("default")
    configuration = system.create_configuration("default")
    atoms = configuration.atoms
    rng = numpy.random.default_rng()
    atno = rng.integers(1, 101, size=natoms)
    x = rng.uniform(low=0, high=100, size=natoms)
    y = rng.uniform(low=0, high=100, size=natoms)
    z = rng.uniform(low=0, high=100, size=natoms)
    t1 = time.perf_counter()
    print(f"      setup took {t1-t0:.3} s")

    t0 = time.perf_counter()
    atoms.append(atno=atno.tolist(), x=x.tolist(), y=y.tolist(), z=z.tolist())
    t1 = time.perf_counter()

    print(f"  appending took {t1-t0:.3} s")
    filename = disk_db.filename
    size = os.path.getsize(filename) / 1024 / 1024
    print(f"       database is {size:.3} MB")

    assert atoms.n_atoms == natoms


@pytest.mark.timing
def test_access(disk_db):
    """Check that we can access the created system"""
    t0 = time.perf_counter()
    atoms = disk_db.system.configuration.atoms
    t1 = time.perf_counter()
    print(f"  test took {t1-t0:.3} s")
    assert atoms.n_atoms == natoms


@pytest.mark.timing
def test_loop_over_atoms(disk_db):
    """Loop over all the atoms (rows)."""
    t0 = time.perf_counter()
    atoms = disk_db.system.configuration.atoms
    n = 0
    for atom in atoms.atoms():
        n += 1
    t1 = time.perf_counter()
    print(f"  test took {t1-t0:.3} s")
    assert n == atoms.n_atoms


@pytest.mark.timing
def test_selection(disk_db):
    """Test selecting a subset of atoms."""
    t0 = time.perf_counter()
    atoms = disk_db.system.configuration.atoms
    n = 0
    for atom in atoms.atoms("atno", "==", 6):
        n += 1
    t1 = time.perf_counter()
    print(f"  test took {t1-t0:.3} s")
    print(f"\nFound {n} carbon atoms in {natoms} atoms")
    print(f"There should be about {int(natoms / 100)}")
    assert atom["atno"] == 6


@pytest.mark.timing
def test_selection_count(disk_db):
    """Test counting a subset of atoms."""
    t0 = time.perf_counter()
    atoms = disk_db.system.configuration.atoms
    n = atoms.get_n_atoms("atno", "==", 6)
    t1 = time.perf_counter()
    print(f"  test took {t1-t0:.3} s")
    print(f"\nFound {n} carbon atoms in {natoms} atoms")
    print(f"There should be about {int(natoms / 100)}")


@pytest.mark.timing
def test_creation_mem(mem_db):
    """Create a system with a million atoms"""
    print(f"\nCreating a system with {natoms} atoms")
    t0 = time.perf_counter()
    system = mem_db.create_system("default")
    configuration = system.create_configuration("default")
    atoms = configuration.atoms
    rng = numpy.random.default_rng()
    atno = rng.integers(1, 101, size=natoms)
    x = rng.uniform(low=0, high=100, size=natoms)
    y = rng.uniform(low=0, high=100, size=natoms)
    z = rng.uniform(low=0, high=100, size=natoms)
    t1 = time.perf_counter()
    print(f"      setup took {t1-t0:.3} s")

    t0 = time.perf_counter()
    atoms.append(atno=atno.tolist(), x=x.tolist(), y=y.tolist(), z=z.tolist())
    t1 = time.perf_counter()

    print(f"  appending took {t1-t0:.3} s")

    assert atoms.n_atoms == natoms


@pytest.mark.timing
def test_access_mem(mem_db):
    """Check that we can access the created system"""
    t0 = time.perf_counter()
    atoms = mem_db.system.configuration.atoms
    t1 = time.perf_counter()
    print(f"  test took {t1-t0:.3} s")
    assert atoms.n_atoms == natoms


@pytest.mark.timing
def test_loop_over_atoms_mem(mem_db):
    """Loop over all the atoms (rows)."""
    t0 = time.perf_counter()
    atoms = mem_db.system.configuration.atoms
    n = 0
    for atom in atoms.atoms():
        n += 1
    t1 = time.perf_counter()
    print(f"  test took {t1-t0:.3} s")
    assert n == atoms.n_atoms


@pytest.mark.timing
def test_selection_mem(mem_db):
    """Test selecting a subset of atoms."""
    t0 = time.perf_counter()
    atoms = mem_db.system.configuration.atoms
    n = 0
    for atom in atoms.atoms("atno", "==", 6):
        n += 1
    t1 = time.perf_counter()
    print(f"  test took {t1-t0:.3} s")
    print(f"\nFound {n} carbon atoms in {natoms} atoms")
    print(f"There should be about {int(natoms / 100)}")
    assert atom["atno"] == 6


@pytest.mark.timing
def test_selection_count_mem(mem_db):
    """Test counting a subset of atoms."""
    t0 = time.perf_counter()
    atoms = mem_db.system.configuration.atoms
    n = atoms.get_n_atoms("atno", "==", 6)
    t1 = time.perf_counter()
    print(f"  test took {t1-t0:.3} s")
    print(f"\nFound {n} carbon atoms in {natoms} atoms")
    print(f"There should be about {int(natoms / 100)}")



# -*- coding: utf-8 -*-

"""A dictionary-like object for holding a system
"""

import collections.abc
import logging
import sqlite3

import molsystem
from .cif import CIFMixin
from .configuration import _Configuration
from .properties import _Properties
from .system import _System
from .table import _Table
from .templates import _Templates

logger = logging.getLogger(__name__)


class SystemDB(CIFMixin, collections.abc.MutableMapping):
    """A database of systems for SEAMM.

    A class based on a SQLite database for describing molecular and
    periodic systems.

    See the documentation at https://molssi-seamm.github.io/molsystem
    for a complete description of the database schema, including a
    diagram.

    The key concepts are:

    System
        The overall container for the configurations, atoms, bonds,
        etc. A system has one or more configurations (conformers),
        each of which details the atoms, bonds, etc.

    Configuration
        The configuration is a single instance of the system, with
        atoms, coordinates, bonds and subsets. It is close to what
        most programs consider the molecule or crystal.

        There may be different set of atoms in different
        configurations of a single system. This supports e.g. grand
        canonical ensembles.

        The set of bonds may also differ between different
        configurations, whether or not the set of atoms changes. This
        supports e.g. reactive forcefields.

        Each configuration also has its own set of subsets, which are
        a way of collecting atoms into groups, and can be thought of
        as a generalization of the chain and residue nomenclature for
        proteins. A configuration may have any number of subsets, and
        atoms may be in more than one subset.

    Atoms
        Each configuration contains a set of atoms, though different
        configurations may contain different atoms. Each atom is
        identified by its atomic number and unique name and has
        coordinates. It may also have other attributes, but that
        depends on the simulation.

    Bonds
        Each configuration may also have information on bonds between
        the atoms. A bond connects two atoms, and has a bond order
        (single, double, triple, aromatic, ...). In periodic systems
        with an infinite network of bonds, each bond also has a cell
        offset to identify the relative cells of the two atoms.

    Templates and Subsets
        Subsets define groups of atoms in a general way. They are
        defined by a template, which may be nothing more than a type
        and name, e.g. 'residue/ala', or the template may be linked
        with a system which has atoms, bonds, etc.

        Each subset is linked to its group atoms. For example, the
        template 'residue/ala' mentioned above would have a subset for
        each alanine in the protein. If the template were linked to a
        system which was the alanine residue, then each atom would
        also be connected with the appropriate atom in the template
        system, so even if the order and names of atoms in the system
        were different, we could still identify each atom with the
        corresponding atom in the template system.

    The tables that implement this are:

    system
        The list of all the systems in the database

    configuration
        The list of all configurations of all systems, labeled by the
        system they belong to.

    atom
        The list of atoms in all systems and configurations. The
        attributes of the atoms do not depend on the configuration,
        i.e. are unchanging.

    bond
        The list of all bonds in all systems and configurations,
        giving the two atoms that are bonded plus the bond order.

    coordinates
        The fractional or Cartesian coordinates of the atoms as well
        as any other atomic properties that vary by configuration.

    subset
        The instances of subsets.

    element
        The periodic table, used as a foreign key to identify atoms.

    template
        The templates -- labels -- for subsets, which may connect the
        subset to a template system.

    configuration_subset
        A joining table used to define which subsets are "in" a
        configuration.

    subset_atom
        A joining table to define with atoms are in a subset. There is
        an optional field for the template atom if the template has an
        associated system. In this case, the template atom identifies
        which atom in the template system is the same as the given
        atom.

    symmetry
        Information about point or space group symmetry.

    symmetryoperation
        The detailed symmetry operations for the point or space
        groups.

    cell
        The information about the periodicity of a configuration, if
        needed.

    atomset
        A set of atoms, used with a joining table to connect atoms
        with configurations.

    atomset_atom
        The joining table connecting atomsets with their atoms.

    bondset
        A set of bonds, used with a joining table to connect bonds
        with configurations.

    bondset_bond
        The joining table connecting bondsets with their bonds.

    :meta public:
    """

    def __init__(self, parent=None, logger=logger, **kwargs):
        self._parent = parent
        self.logger = logger
        self._attached = {}
        self._current_system_id = None  # The id of the current system
        self._checkpoints = []
        self._filename = None
        self._db = None
        self._cursor = None
        self._items = {}

        if "filename" in kwargs:
            self.filename = kwargs.pop("filename")
        else:
            self.filename = "seamm.db"

    def __del__(self):
        """Destructor: need to close the database if any."""

        # Delete any cached objects
        del self._items

        # And close the database
        if self._db is not None:
            self.db.commit()
            self.db.close()

    def __enter__(self):
        self.db.commit()

        backup = self.parent.copy_system(self, temporary=True)
        self._checkpoints.append(backup)
        return self

    def __exit__(self, etype, value, traceback):
        backup = self._checkpoints.pop()

        if etype is None:
            self.db.commit()

            # Log the changes
            diffs = self.diff(backup)
            if len(diffs) > 0:
                self.version = self.version + 1

            # Not sure why this commit is needed...
            self.db.commit()
        else:
            self.parent.overwrite(self, backup)

        # and delete the copy
        del self.parent[backup.nickname]

    def __getitem__(self, key):
        """Allow [] access to the dictionary!

        Because some of the items, such as template contain state,
        we need to ensure that the same object is used everywhere. Hence
        the self._items array to store the instances.

        Parameters
        ----------
        key : str
            The table name or name of a multi-table item like atoms.

        Returns
        -------
        table : Table
        """
        if key not in self._items:
            self._items[key] = _Table(self, key)
        return self._items[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        raise NotImplementedError(f"Table '{key}' cannot be created yet")

    def __delitem__(self, key):
        """Allow deletion of keys"""
        if key in self:
            self.cursor.execute(f"DROP TABLE '{key}'")

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self.list())

    def __len__(self):
        """The len() command"""
        self.cursor.execute(
            "SELECT COUNT(*)" "  FROM sqlite_master" " WHERE type = 'table'"
        )
        return self.cursor.fetchone()[0]

    # def __repr__(self):
    #     """The string representation of this object"""
    #     raise NotImplementedError()

    # def __str__(self):
    #     """The pretty string representation of this object"""
    #     raise NotImplementedError()

    def __contains__(self, table):
        """Return a boolean indicating if a key exists."""
        # Normal the tablename is used as an identifier, so is quoted with ".
        # Here we need it as a string literal so strip any quotes from it.
        if "." in table:
            schema, table = table.split(".")
            schema = schema.strip('"')
        else:
            schema = "main"

        table = table.strip('"')
        self.cursor.execute(
            "SELECT COUNT(*)"
            f" FROM {schema}.sqlite_master"
            f" WHERE type = 'table' and name = '{table}'"
        )
        return self.cursor.fetchone()[0] == 1

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        # LGTM doesn't like using 'is', but this is equivalent.
        if self.id() == other.id():
            return True

        tables = set(self.list())
        other_tables = set(other.list())

        added = tables - other_tables
        deleted = other_tables - tables
        in_common = tables & other_tables

        if len(added) > 0:
            return False
        if len(deleted) > 0:
            return False

        # Need the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.
        name = self.name
        other_name = other.name
        detach = False
        if name != other_name and not self.is_attached(other):
            # Attach the other system in order to do comparisons.
            self.attach(other)
            detach = True

        # Check the tables in both systems
        result = True
        for table in in_common:
            if self[table] != other[table]:
                result = False
                break

        # Detach the other database if needed
        if detach:
            self.detach(other)

        return result

    @property
    def cursor(self):
        """A database cursor."""
        return self._cursor

    @property
    def db(self):
        """The database connection."""
        return self._db

    @property
    def db_version(self):
        """The version string for the database."""
        self.cursor.execute("SELECT value FROM metadata WHERE key = 'version'")
        return self.cursor.fetchone()[0]

    @property
    def filename(self):
        """The name of the file (or URI) for the database."""
        return self._filename

    @filename.setter
    def filename(self, value):
        if value != self._filename:
            if self._db is not None:
                self.cursor.close()
                self._db.commit()
                self._db.close()
                self._db = None
                self._cursor = None
        self._filename = value
        if self._filename is not None:
            if self._filename[0:5] == "file:":
                self._db = sqlite3.connect(self._filename, uri=True)
            else:
                self._db = sqlite3.connect(self._filename)
            self._db.row_factory = sqlite3.Row
            self._db.execute("PRAGMA foreign_keys = ON")
            self._cursor = self._db.cursor()
            self._initialize()

    @property
    def names(self):
        """The names of the system."""
        return [row[0] for row in self.db.execute("SELECT name FROM system")]

    @property
    def n_systems(self):
        """The number of systems in the database."""
        return self["system"].n_rows

    @property
    def n_templates(self):
        """The number of templates."""
        return self.templates.n_rows

    @property
    def parent(self):
        """The parent of this, i.e. a Systems object."""
        return self._parent

    @property
    def properties(self):
        """The class to handle the properties."""
        return _Properties(self)

    @property
    def system(self):
        """The current system object."""
        if self.n_systems == 0:
            return None
        if self._current_system_id is None:
            self._current_system_id = self.system_ids[-1]
        return _System(self, self._current_system_id)

    @system.setter
    def system(self, value):
        # Might be a system or an integer
        if isinstance(value, _System):
            value = value.id

        if value not in self.system_ids:
            raise KeyError(f"system '{value}' does not exist.")
        self._current_system_id = value

    @property
    def systems(self):
        """The list of system objects."""
        return [_System(self, _id) for _id in self.system_ids]

    @property
    def system_ids(self):
        """The list of system ids."""
        return [row[0] for row in self.db.execute("SELECT id FROM system")]

    @property
    def templates(self):
        """The defined templates."""
        return _Templates(self)

    def attach(self, other):
        """Attach another system to this one's database.

        Parameters
        ----------
        other : SystemDB
            The other SystemDB object containing the database

        Returns
        -------
        name : str
            The attachment name.
        """
        if not self.is_attached(other):
            # Get a unique name for attaching as
            n = len(self._attached)
            attached_name = f"db_{n}"
            while attached_name in self._attached.values():
                n += 1
                attached_name = f"db_{n}"

            # and attach
            self.db.execute(f"ATTACH DATABASE '{other.filename}' AS '{attached_name}'")
            self._attached[other.filename] = attached_name
        return self._attached[other.filename]

    def attached_as(self, other):
        """The attachment name for another system.

        Parameters
        ----------
        other : SystemDB
            The other SystemDB object containing the attached database

        Returns
        -------
        name : str
            The attachment name.
        """
        return self._attached[other.filename]

    def attributes(self, tablename: str):
        """The attributes -- columns -- of a given table.

        Parameters
        ----------
        tablename : str
            The name of the table, optionally including the schema followed by
            a dot.

        Returns
        -------
        attributes : Dict[str, Any]
            A dictionary of dictionaries for the attributes and their
            descriptors
        """
        if "." in tablename:
            schema, tablename = tablename.split(".")
            sql = f"PRAGMA {schema}.table_info('{tablename}')"
        else:
            sql = f"PRAGMA table_info('{tablename}')"

        result = {}
        for line in self.db.execute(sql):
            result[line["name"]] = {
                "type": line["type"],
                "notnull": bool(line["notnull"]),
                "default": line["dflt_value"],
                "primary key": bool(line["pk"]),
            }
        return result

    def close(self):
        """Close the database."""
        self.filename = None

    def create_system(self, name="", make_current=True):
        """Add a new system.

        Parameters
        ----------
        name : str = None
            A user-friendly name for the system, defaults to no name.
        make_current : bool = True
            If True, make this the current system.

        Returns
        -------
        _System
            The newly created system.
        """

        _id = self["system"].append(name=name)[0]

        if make_current:
            self._current_system_id = _id

        return _System(self, _id)

    def create_table(self, name, cls=_Table, other=None):
        """Create a new table with the given name.

        Parameters
        ----------
        name : str
            The name of the new table.

        cls : Table subclass
            The class of the new table, defaults to Table

        Returns
        -------
        table : class Table
            The new table
        """
        if name in self:
            raise KeyError(f"'{name}' already exists in the system.")

        self._items[name] = cls(self, name, other)
        return self._items[name]

    def delete_system(self, system):
        """Delete an existing system.

        Parameters
        ----------
        system : int or _System
            The system to delete.

        Returns
        -------
        None
        """

        if isinstance(system, _System):
            sid = system.id
        else:
            sid = system

        sql = "DELETE FROM system WHERE id = ?"
        self.db.execute(sql, (sid,))

        if self._current_system_id == sid:
            self._current_system_id = None

    def detach(self, other):
        """Detach an attached system.

        Parameters
        ----------
        other : SystemDB
            The other SystemDB object containing the database
        """
        if self.is_attached(other):
            attached_name = self.attached_as(other)
            self.cursor.execute(f'DETACH DATABASE "{attached_name}"')
            del self._attached[other.filename]

    def diff(self, other):
        """Differences between this system and another."""
        result = {}

        if self.id() == other.id():
            return result

        tables = set(self.list())
        other_tables = set(other.list())

        added = tables - other_tables
        deleted = other_tables - tables
        in_common = tables & other_tables

        if len(added) > 0:
            result["tables added"] = added
        if len(deleted) > 0:
            result["tables deleted"] = deleted

        # Need the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.
        name = self.filename
        other_name = other.filename
        detach = False
        if name != other_name and not self.is_attached(other):
            # Attach the other system in order to do comparisons.
            self.attach(other)
            detach = True

        # Check the tables in both systems
        for table in in_common:
            tmp = self[table].diff(other[table])
            if len(tmp) > 0:
                result[f"table '{table}' diffs"] = tmp

        # Detach the other database if needed
        if detach:
            self.detach(other)

        return result

    def find_configurations(self, atomset=None, bondset=None):
        """Return the configurations that have given atom- or bondsets

        Parameters
        ----------
        atomset : int = None
            The id of the atomset.
        bondset : int = None
            The id of the bondset.

        Returns
        -------
        [_Configuration]
        """

        if atomset is not None:
            if bondset is not None:
                self.cursor.execute(
                    "SELECT id FROM configuration WHERE atomset = ? AND bondset = ?",
                    (atomset, bondset),
                )
            else:
                self.cursor.execute(
                    "SELECT id FROM configuration WHERE atomset = ?", (atomset,)
                )
            return [
                _Configuration(_id=cid, system_db=self)
                for cid in self.cursor.fetchall()
            ]
        elif bondset is not None:
            self.cursor.execute(
                "SELECT id FROM configuration WHERE bondset = ?", (bondset,)
            )
            return [
                _Configuration(_id=cid, system_db=self)
                for cid in self.cursor.fetchall()
            ]
        else:
            raise RuntimeError("Must give atomset or bondset.")

    def get_configuration(self, cid):
        """Return the specified configuration.

        Parameters
        ----------
        cid : int
            The id of the configuration.

        Returns
        -------
        _Configuration
            The requested configuration.
        """
        return _Configuration(_id=cid, system_db=self)

    def get_system(self, id_or_name):
        """Get the specified system object.

        Parameters
        ----------
        id_or_name : int or str
            The id (int) or name (str) of the system

        Returns
        -------
        _System
            The requested system.

        Raises
        ------
        ValueError
            If the system does not exist, or more than one have the requested
            name.
        """
        if isinstance(id_or_name, str):
            # The name of the system ... find it.
            sql = "SELECT id FROM SYSTEM WHERE name = ?"
            systems = [x[0] for x in self.db.execute(sql, (id_or_name,))]
            if len(systems) == 0:
                raise ValueError(f"The system '{id_or_name}' does not exist.")
            elif len(systems) > 1:
                raise ValueError(f"There is more than one system named '{id_or_name}'")
            else:
                id_or_name = systems[0]
        return _System(self, id_or_name)

    def is_attached(self, other):
        """Return whether another system is attached to this one.

        Parameters
        ----------
        other : SystemDB
            The other SystemDB object containing the database

        Returns
        -------
        bool
            Whether the database is attached.
        """
        return other.filename in self._attached

    def list(self):
        """Return a list of all the tables in the system."""
        result = []
        for row in self.db.execute(
            "SELECT name" "  FROM sqlite_master" " WHERE type = 'table'"
        ):
            result.append(row["name"])
        return result

    def _initialize(self):
        """Initialize the SQLite database.

        The order is a bit tricky, since many tables reference other tables,
        and hence need to be created after the ones that they reference.

        Start with the standalone tables:
        """

        # If the database is initialized, the metadata table exists.
        # In the future we might need to check the version and upgrade
        # older versions, but now at version 1.0 we are all done!
        if "metadata" not in self:
            # metadata, where we store, get the database version
            table = self["metadata"]
            table.add_attribute("key", coltype="str", pk=True)
            table.add_attribute("value", coltype="str")

            table.append(key="version", value="1.0")
            self.db.commit()

            # The element table
            table = self["element"]
            table.add_attribute("atno", coltype="int", pk=True)
            table.add_attribute("symbol", coltype="str", index="unique")
            table.add_attribute("mass", coltype="float")

            # and fill it from the element data
            for symbol, data in molsystem.elements.data.items():
                table.append(
                    symbol=symbol,
                    atno=data["atomic number"],
                    mass=data["atomic weight"],
                )
            self.db.commit()

            # Symmetry information
            table = self["symmetry"]
            table.add_attribute("id", coltype="int", pk=True)
            table.add_attribute("group", coltype="str")

            table = self["symmetryoperation"]
            table.add_attribute("symmetry", coltype="int", references="symmetry")
            table.add_attribute("symop", coltype="str")

            # Periodic cell information
            table = self["cell"]
            table.add_attribute("id", coltype="int", pk=True)
            table.add_attribute("a", coltype="float", default=10.0)
            table.add_attribute("b", coltype="float", default=10.0)
            table.add_attribute("c", coltype="float", default=10.0)
            table.add_attribute("alpha", coltype="float", default=90.0)
            table.add_attribute("beta", coltype="float", default=90.0)
            table.add_attribute("gamma", coltype="float", default=90.0)

            # The systems themselves
            table = self["system"]
            table.add_attribute("id", coltype="int", pk=True)
            table.add_attribute("name", coltype="str", notnull=True, default="default")

            # The atoms, and the sets of atoms
            table = self["atom"]
            table.add_attribute("id", coltype="int", pk=True)
            table.add_attribute("atno", coltype="int", references="element")

            table = self["atomset"]
            table.add_attribute("id", coltype="int", pk=True)

            table = self["atomset_atom"]
            table.add_attribute("atomset", coltype="int", references="atomset")
            table.add_attribute("atom", coltype="int", references="atom")
            self.db.execute(
                "CREATE INDEX 'idx_atomset_atom_atomset_atom' "
                'ON atomset_atom ("atomset", "atom")'
            )

            # The bonds, and sets of bonds
            table = self["bond"]
            table.add_attribute("id", coltype="int", pk=True)
            table.add_attribute("i", coltype="int", references="atom")
            table.add_attribute("j", coltype="int", references="atom")
            table.add_attribute("bondorder", coltype="int", default=1)
            # Only need for infinite networked systems. Think about later!
            # table.add_attribute("offset1", coltype="int", default=0)
            # table.add_attribute("offset2", coltype="int", default=0)
            # table.add_attribute("offset3", coltype="int", default=0)

            table = self["bondset"]
            table.add_attribute("id", coltype="int", pk=True)

            table = self["bondset_bond"]
            table.add_attribute("bondset", coltype="int", references="bondset")
            table.add_attribute("bond", coltype="int", references="bond")
            self.db.execute(
                "CREATE INDEX 'idx_bondset_bond_bondset_bond' "
                'ON bondset_bond ("bondset", "bond")'
            )

            # Now we can set up the configurations
            table = self["configuration"]
            table.add_attribute("id", coltype="int", pk=True)
            table.add_attribute("name", coltype="str")
            table.add_attribute("system", coltype="int", references="system")
            table.add_attribute("version", coltype="int", notnull=True, default=0)
            table.add_attribute("periodicity", coltype="int", default=0)
            table.add_attribute("coordinate_system", coltype="str", default="Cartesian")
            table.add_attribute("symmetry", coltype="int", references="symmetry")
            table.add_attribute("cell", coltype="int", references="cell")
            table.add_attribute("atomset", coltype="int", references="atomset")
            table.add_attribute("bondset", coltype="int", references="bondset")
            table.add_attribute("charge", coltype="int", notnull=True, default=0)
            table.add_attribute("spin_multiplicity", coltype="int")
            table.add_attribute("n_active_electrons", coltype="int", default=0)
            table.add_attribute("n_active_orbitals", coltype="int", default=0)
            table.add_attribute("state", coltype="str", default="1")

            # And coordinates, which depend on configurations
            table = self["coordinates"]
            table.add_attribute(
                "configuration", coltype="int", references="configuration"
            )
            table.add_attribute("atom", coltype="int", references="atom")
            table.add_attribute("x", coltype="float")
            table.add_attribute("y", coltype="float")
            table.add_attribute("z", coltype="float")
            self.db.execute(
                "CREATE INDEX 'idx_coordinates_atom' ON coordinates (\"atom\")"
            )
            self.db.execute(
                "CREATE INDEX idx_coordinates_atom_configuration "
                "    ON coordinates(atom, configuration)"
            )
            # The definition of the subsets -- templates
            table = self["template"]
            table.add_attribute("id", coltype="int", pk=True)
            table.add_attribute("name", coltype="str")
            table.add_attribute("canonical_smiles", coltype="str")
            table.add_attribute("category", coltype="str", default="general")
            table.add_attribute(
                "configuration", coltype="int", references="configuration"
            )
            self.db.execute(
                "CREATE UNIQUE INDEX 'idx_template_name_type'"
                '    ON template ("name", "type")'
            )

            # The subsets
            table = self["subset"]
            table.add_attribute("id", coltype="int", pk=True)
            table.add_attribute(
                "configuration", coltype="int", references="configuration"
            )
            table.add_attribute("template", coltype="int", references="template")
            self.db.execute(
                "CREATE INDEX subset_idx_template_configuration "
                "ON subset(template, configuration)"
            )

            # The connection between subsets and the atoms in the system
            table = self["subset_atom"]
            table.add_attribute("atom", coltype="int", references="atom")
            table.add_attribute("subset", coltype="int", references="subset")
            table.add_attribute("templateatom", coltype="int", references="atom")
            self.db.execute(
                "CREATE INDEX 'idx_subset_atom_subset_atom' "
                'ON subset_atom ("subset", "atom")'
            )

            #####################################################
            # The star schema for storing properties and features
            #####################################################

            properties = self.properties
            properties.create_schema()

            self.db.commit()

    def system_exists(self, id_or_name):
        """See if the given system exists.

        Parameters
        ----------
        id_or_name : int or str
            The id (int) or name (str) of the system

        Returns
        -------
        bool
            Whether it exists.
        """
        if isinstance(id_or_name, str):
            # The name of the system ... find it.
            sql = "SELECT COUNT(*) FROM SYSTEM WHERE name = ?"
            self.cursor.execute(sql, (id_or_name,))
        else:
            sql = "SELECT COUNT(*) FROM SYSTEM WHERE id = ?"
            self.cursor.execute(sql, (id_or_name,))

        return self.cursor.fetchone()[0] != 0
